# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import argparse
import math
import warnings
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy
from torchvision.ops.misc import SqueezeExcitation as SElayer

from ..op.registry import ACTIVATION_REGISTRY, NORM_REGISTRY, POOL2D_RESGISTRY


def make_divisible(
    v: Union[float, int],
    divisor: Optional[int] = 8,
    min_value: Optional[Union[float, int]] = None,
) -> Union[float, int]:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

def auto_pad(kernel_size: Union[int, Tuple[int, int]], dilation: int = 1, **kwargs) -> Tuple[int, int]:
    """
    Auto Padding for the convolution blocks
    """
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    if isinstance(dilation, int):
        dilation = (dilation, dilation)

    pad_h = ((kernel_size[0] - 1) * dilation[0]) // 2
    pad_w = ((kernel_size[1] - 1) * dilation[1]) // 2
    return (pad_h, pad_w)


class ConvLayer(nn.Module):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        padding: Optional[Union[int, Tuple[int, int]]] = None,
        groups: Optional[int] = 1,
        bias: bool = False,
        padding_mode: Optional[str] = 'zeros',
        use_norm: bool = True,
        norm_type: Optional[str] = None,
        use_act: bool = True,
        act_type: Optional[str] = None,
    ) -> None:
        super().__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)

        if isinstance(stride, int):
            stride = (stride, stride)

        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        assert isinstance(kernel_size, Tuple)
        assert isinstance(stride, Tuple)
        assert isinstance(dilation, Tuple)

        if padding is None:
            padding = (
                int((kernel_size[0] - 1) / 2) * dilation[0],
                int((kernel_size[1] - 1) / 2) * dilation[1],
            )

        block = nn.Sequential()

        conv_layer = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )

        block.add_module(name='conv', module=conv_layer)

        self.norm_name = None
        if use_norm or norm_type is not None:
            _norm_type = norm_type.lower() if norm_type is not None else 'batch_norm'
            if bias:
                warnings.warn("Bias would be ignored in batch normalization", stacklevel=2)
            assert _norm_type in NORM_REGISTRY
            cls_norm = NORM_REGISTRY[_norm_type]
            norm_layer = cls_norm(num_features=out_channels)
            block.add_module(name='norm', module=norm_layer)
            self.norm_name = norm_layer.__class__.__name__

        self.act_name = None
        if use_act or act_type is not None:
            _act_type = act_type.lower() if act_type is not None else 'relu'
            assert _act_type in ACTIVATION_REGISTRY
            cls_act = ACTIVATION_REGISTRY[_act_type]
            act_layer = cls_act()
            block.add_module(name='act', module=act_layer)
            self.act_name = act_layer.__class__.__name__

        self.block = block

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        return self.block(x)

    def __repr__(self):
        return f"{self.block}"


class SeparableConvLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        padding: Optional[Union[int, Tuple[int, int]]] = None,
        bias: bool = False,
        padding_mode: Optional[str] = 'zeros',
        use_norm: bool = True,
        norm_type: Optional[str] = None,
        use_act: bool = True,
        act_type: Optional[str] = None,
        no_out_act: Optional[bool] = False,
    ) -> None:
        super().__init__()
        if act_type is None:
            act_type = 'relu'
        self.depthwise = ConvLayer(in_channels=in_channels, out_channels=in_channels,
                                   kernel_size=kernel_size, stride=stride, dilation=dilation,
                                   padding=padding, groups=in_channels, bias=bias, padding_mode=padding_mode,
                                   use_norm=use_norm, norm_type=norm_type, use_act=use_act, act_type=act_type,)
        self.pointwise = ConvLayer(in_channels=in_channels, out_channels=out_channels, kernel_size=1,
                                   use_norm=use_norm, norm_type=norm_type, use_act=False)
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        x = self.depthwise(x)
        x = self.pointwise(x)
        x = self.final_act(x)
        return x

class RepVGGBlock(nn.Module):
    """
    A convolutional block that combines two convolution layers (kernel and point-wise conv).
    This implementation is based on https://github.com/lyuwenyu/RT-DETR/blob/b444daf79cf25f95b740ae71e80fd165e892739a/rtdetr_pytorch/src/zoo/rtdetr/hybrid_encoder.py#L35.
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: Union[int, Tuple[int, int]] = 3,
                 groups: int = 1,
                 act_type: Optional[str] = None,
                 use_identity: Optional[bool]=True,
                 ):
        if act_type is None:
            act_type = 'silu'
        super().__init__()
        assert isinstance(in_channels, int)
        assert isinstance(out_channels, int)
        assert isinstance(act_type, str)
        assert kernel_size == 3

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups
        self.conv1 = ConvLayer(in_channels, out_channels, kernel_size, groups=groups, use_act=False)
        self.conv2 = ConvLayer(in_channels, out_channels, 1, groups=groups, use_act=False)
        self.rbr_identity = nn.BatchNorm2d(num_features=in_channels) if use_identity and out_channels == in_channels else None

        assert act_type in ACTIVATION_REGISTRY
        self.act = ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        if hasattr(self, 'conv'):
            y = self.conv(x)
            return y

        y = self.conv1(x) + self.conv2(x) + self.rbr_identity(x) if self.rbr_identity else self.conv1(x) + self.conv2(x)

        return self.act(y)

    def convert_to_deploy(self):
        if not hasattr(self, 'conv'):
            self.conv = nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, stride=1, padding=1)

        kernel, bias = self.get_equivalent_kernel_bias()
        self.conv.weight.data = kernel
        self.conv.bias.data = bias
        # self.__delattr__('conv1')
        # self.__delattr__('conv2')

    def get_equivalent_kernel_bias(self):
        kernel3x3, bias3x3 = self._fuse_bn_tensor(self.conv1)
        kernel1x1, bias1x1 = self._fuse_bn_tensor(self.conv2)
        if self.rbr_identity:
            kernelid, biasid = self._fuse_bn_tensor(self.rbr_identity)
        else:
            kernelid = 0
            biasid = 0

        return kernel3x3 + self._pad_1x1_to_3x3_tensor(kernel1x1) + kernelid, bias3x3 + bias1x1 + biasid

    def _pad_1x1_to_3x3_tensor(self, kernel1x1):
        if kernel1x1 is None:
            return 0
        else:
            return nn.functional.pad(kernel1x1, [1, 1, 1, 1])

    def _fuse_bn_tensor(self, branch: Union[ConvLayer, nn.BatchNorm2d, nn.Identity]):
        if isinstance(branch, nn.Identity):
            return 0, 0
        if isinstance(branch, nn.BatchNorm2d):
            if not hasattr(self, 'id_tensor'):
                input_dim = self.in_channels // self.groups
                kernel_value = np.zeros((self.in_channels, input_dim, 3, 3), dtype=np.float32)
                for i in range(self.in_channels):
                    kernel_value[i, i % input_dim, 1, 1] = 1
                self.id_tensor = torch.from_numpy(kernel_value).to(branch.weight.device)
            kernel = self.id_tensor
            running_mean = branch.running_mean
            running_var = branch.running_var
            gamma = branch.weight
            beta = branch.bias
            eps = branch.eps
            std = (running_var + eps).sqrt()
            t = (gamma / std).reshape(-1, 1, 1, 1)
        else:
            assert isinstance(branch, ConvLayer)
            kernel = branch.block.conv.weight
            running_mean = branch.block.norm.running_mean
            running_var = branch.block.norm.running_var
            gamma = branch.block.norm.weight
            beta = branch.block.norm.bias
            eps = branch.block.norm.eps
            std = (running_var + eps).sqrt()
            t = (gamma / std).reshape(-1, 1, 1, 1)
        return kernel * t, beta - running_mean * gamma / std


class Pool2d(nn.Module):
    def __init__(self,
                 method: str = "max",
                 kernel_size: int = 2,
                 stride: Optional[int] = None,
                 padding: int = 0,
                 **kwargs):
        super().__init__()
        assert method.lower() in POOL2D_RESGISTRY
        self.pool = POOL2D_RESGISTRY[method.lower()](kernel_size=kernel_size, stride=stride, padding=padding, **kwargs)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        return self.pool(x)


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[str] = None,
        expansion: Optional[int] = None,
        act_type: Optional[str] = None,
        no_out_act: bool = False
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = 'batch_norm'
        if act_type is None:
            act_type = 'relu'
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        if expansion is not None:
            self.expansion = expansion
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1

        self.conv1 = ConvLayer(in_channels=inplanes, out_channels=planes,
                               kernel_size=3, stride=stride, dilation=1, padding=1, groups=1,
                               norm_type=norm_layer, act_type=act_type)

        self.conv2 = ConvLayer(in_channels=planes, out_channels=planes,
                               kernel_size=3, stride=1, dilation=1, padding=1, groups=1,
                               norm_type=norm_layer, use_act=False)

        self.downsample = downsample
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.final_act(out)

        return out


class Bottleneck(nn.Module):
    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[str] = None,
        expansion: Optional[int] = None,
        act_type: Optional[str] = None,
        no_out_act: bool = False
    ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = 'batch_norm'
        if act_type is None:
            act_type = 'relu'
        width = int(planes * (base_width / 64.)) * groups
        if expansion is not None:
            self.expansion = expansion
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1

        self.conv1 = ConvLayer(in_channels=inplanes, out_channels=width,
                               kernel_size=1, stride=1,
                               norm_type=norm_layer, act_type=act_type)

        self.conv2 = ConvLayer(in_channels=width, out_channels=width,
                               kernel_size=3, stride=stride, dilation=dilation, padding=dilation, groups=groups,
                               norm_type=norm_layer, act_type=act_type)

        self.conv3 = ConvLayer(in_channels=width, out_channels=planes * self.expansion,
                               kernel_size=1, stride=1,
                               norm_type=norm_layer, use_act=False)

        self.downsample = downsample
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.final_act(out)

        return out


class RepNBottleneck(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 shortcut: bool = True,
                 expansion: float = 1.0,
                 depthwise: bool = False,
                 act_type: Optional[str] = None,
                 **kwargs):
        super().__init__()
        hidden_channels = int(out_channels * expansion)
        self.conv1 = RepVGGBlock(in_channels, hidden_channels, 3, act_type=act_type, **kwargs)
        if depthwise:
            self.conv2 = SeparableConvLayer(hidden_channels,
                                            out_channels,
                                            kernel_size=3, stride=1,
                                            act_type=act_type)
        else:
            self.conv2 = ConvLayer(hidden_channels,
                                   out_channels,
                                   kernel_size=3, stride=1,
                                   act_type=act_type)
        self.shortcut = shortcut

        if shortcut and (in_channels != out_channels):
            self.shortcut = False
            warnings.warn(f"Residual connection disabled: in_channels ({in_channels}) != out_channels ({out_channels})", stacklevel=2)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        y = self.conv2(self.conv1(x))
        return x + y if self.shortcut else y


class InvertedResidual(nn.Module):
    # Implemented as described at section 5 of MobileNetV3 paper
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        norm_type: Optional[str] = None,
        act_type: Optional[str] = None,
        use_se: bool = False,
        se_layer: Callable[..., nn.Module] = partial(SElayer, scale_activation=nn.Hardsigmoid),
    ):
        super().__init__()
        if not (1 <= stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = stride == 1 and in_channels == out_channels

        layers: List[nn.Module] = []

        # expand
        if hidden_channels != in_channels:
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=hidden_channels,
                    kernel_size=1,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        # depthwise
        stride = 1 if dilation > 1 else stride
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                groups=hidden_channels,
                norm_type=norm_type,
                act_type=act_type,
            )
        )
        if use_se:
            squeeze_channels = make_divisible(hidden_channels // 4, 8)
            layers.append(se_layer(hidden_channels, squeeze_channels))

        # project
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=out_channels,
                kernel_size=1,
                norm_type=norm_type,
                use_act=False
            )
        )

        self.block = nn.Sequential(*layers)
        self.out_channels = out_channels
        self._is_cn = stride > 1

    def forward(self, input: Tensor) -> Tensor:
        result = self.block(input)
        if self.use_res_connect:
            result = result + input
        return result


class UniversalInvertedResidualBlock(nn.Module):
    # Based on MobileNetV4: https://arxiv.org/pdf/2404.10518
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
        extra_dw: bool = False,
        extra_dw_kernel_size: Optional[Union[int, Tuple[int, int]]] = None,
        middle_dw: bool = True,
        middle_dw_kernel_size: Union[int, Tuple[int, int]] = 3,
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        norm_type: Optional[str] = None,
        act_type: Optional[str] = None,
        use_se: bool = False,
        se_layer: Callable[..., nn.Module] = partial(SElayer, scale_activation=nn.Hardsigmoid),
        layer_scale: Optional[float] = None,
    ):
        super().__init__()
        if not (1 <= stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = stride == 1 and in_channels == out_channels
        layers: List[nn.Module] = []

        # extra depthwise conv
        if extra_dw:
            assert extra_dw_kernel_size is not None, "if extra_dw is True, extra_dw_kernel_size must be provided."
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    kernel_size=extra_dw_kernel_size,
                    stride=1,
                    groups=in_channels,
                    norm_type=norm_type,
                    use_act=False # No activation for extra depthwise conv
                )
            )

        # expand
        if hidden_channels != in_channels:
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=hidden_channels,
                    kernel_size=1,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        # middle depthwise
        if middle_dw:
            assert middle_dw_kernel_size is not None, "if middle_dw is True, middle_dw_kernel_size must be provided."
            layers.append(
                ConvLayer(
                    in_channels=hidden_channels,
                    out_channels=hidden_channels,
                    kernel_size=middle_dw_kernel_size,
                    stride=stride,
                    groups=hidden_channels,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        if use_se:
            squeeze_channels = make_divisible(hidden_channels // 4, 8)
            layers.append(se_layer(hidden_channels, squeeze_channels))

        # project
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=out_channels,
                kernel_size=1,
                norm_type=norm_type,
                use_act=False # No activation for project conv
            )
        )

        self.block = nn.Sequential(*layers)
        self.apply_layer_scale = False
        if layer_scale is not None:
            self.apply_layer_scale = True
            self.layer_scale = LayerScale2d(out_channels, layer_scale)
        self.out_channels = out_channels
        self._is_cn = stride > 1

    def forward(self, input: Tensor) -> Tensor:
        result = self.block(input)
        if self.apply_layer_scale:
            result = self.layer_scale(result)
        if self.use_res_connect:
            result = result + input
        return result


class SinusoidalPositionalEncoding(nn.Module):
    """
    This layer adds sinusoidal positional embeddings to a 3D input tensor. The code has been adapted from
    `Pytorch tutorial <https://pytorch.org/tutorials/beginner/transformer_tutorial.html>`_

    Args:
        d_model (int): dimension of the input tensor
        dropout (Optional[float]): Dropout rate. Default: 0.0
        max_len (Optional[int]): Max. number of patches (or seq. length). Default: 5000
        channels_last (Optional[bool]): Channels dimension is the last in the input tensor

    Shape:
        - Input: :math:`(N, C, P)` or :math:`(N, P, C)` where :math:`N` is the batch size, :math:`C` is the embedding dimension,
            :math:`P` is the number of patches
        - Output: same shape as the input

    """

    def __init__(
        self,
        d_model: int,
        max_len: Optional[int] = 5000,
        channels_last: Optional[bool] = True,
        *args,
        **kwargs
    ) -> None:

        position_last = not channels_last

        pos_encoding = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pos_encoding[:, 0::2] = torch.sin(position * div_term)
        pos_encoding[:, 1::2] = torch.cos(position * div_term)
        # add dummy batch dimension
        pos_encoding = pos_encoding.unsqueeze(0)  # [1 x C x P_max)

        patch_dim = -2  # patch dimension is second last (N, P, C)
        if position_last:
            pos_encoding = pos_encoding.transpose(
                1, 2
            )  # patch dimension is last (N, C, P)
            patch_dim = -1  # patch dimension is last (N, C, P)

        super().__init__()

        self.patch_dim = patch_dim
        self.register_buffer("pe", pos_encoding)

    def forward_patch_last(
        self, x, indices: Optional[Tensor] = None, *args, **kwargs
    ) -> Tensor:
        # seq_length should be the last dim
        if indices is None:
            x = x + self.pe[..., : x.shape[-1]]
        else:
            ndim = x.ndim
            repeat_size = [x.shape[0]] + [-1] * (ndim - 1)

            pe = self.pe.expand(repeat_size)
            selected_pe = torch.gather(pe, index=indices, dim=-1)
            x = x + selected_pe
        return x

    def forward_others(
        self, x, indices: Optional[Tensor] = None, *args, **kwargs
    ) -> Tensor:
        # seq_length should be the second last dim

        # @deepkyu: [fx tracing] Always `indices` is None
        # if indices is None:
        #     x = x + self.pe[..., : x.shape[-2], :]
        # else:
        #     ndim = x.ndim
        #     repeat_size = [x.shape[0]] + [-1] * (ndim - 1)

        #     pe = self.pe.expand(repeat_size)
        #     selected_pe = torch.gather(pe, index=indices, dim=-2)
        #     x = x + selected_pe

        # x = x + self.pe[..., :seq_index, :]
        if not isinstance(x, torch.fx.Proxy):
            self.last_token_num = x.shape[-2]
        x = x + self.pe[..., : self.last_token_num, :]

        return x

    def forward(self, x, indices: Optional[Tensor] = None, *args, **kwargs) -> Tensor:
        if self.patch_dim == -1:
            return self.forward_patch_last(x, indices=indices)
        else:
            return self.forward_others(x, indices=indices)

    # def profile_module(self, input: Tensor) -> Tuple[Tensor, float, float]:
    #     return input, 0.0, 0.0

    # def __repr__(self):
    #     return "{}(dropout={})".format(self.__class__.__name__, self.dropout.p)


class GlobalPool(nn.Module):
    """
    This layers applies global pooling over a 4D or 5D input tensor

    Args:
        pool_type (Optional[str]): Pooling type. It can be mean, rms, or abs. Default: `mean`
        keep_dim (Optional[bool]): Do not squeeze the dimensions of a tensor. Default: `False`

    Shape:
        - Input: :math:`(N, C, H, W)` or :math:`(N, C, D, H, W)`
        - Output: :math:`(N, C, 1, 1)` or :math:`(N, C, 1, 1, 1)` if keep_dim else :math:`(N, C)`
    """

    pool_types = ["mean", "rms", "abs"]

    def __init__(
        self,
        pool_type: Optional[str] = "mean",
        keep_dim: Optional[bool] = False,
        *args,
        **kwargs
    ) -> None:
        super().__init__()
        # if pool_type not in self.pool_types:
        #     logger.error(
        #         "Supported pool types are: {}. Got {}".format(
        #             self.pool_types, pool_type
        #         )
        #     )
        self.pool_type = pool_type
        self.keep_dim = keep_dim

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        cls_name = "{} arguments".format(cls.__name__)
        group = parser.add_argument_group(title=cls_name, description=cls_name)
        group.add_argument(
            "--model.layer.global-pool",
            type=str,
            default="mean",
            help="Which global pooling?",
        )
        return parser

    def _global_pool(self, x: Tensor, dims: List):
        if self.pool_type == "rms":  # root mean square
            x = x**2
            x = torch.mean(x, dim=dims, keepdim=self.keep_dim)
            x = x**-0.5
        elif self.pool_type == "abs":  # absolute
            x = torch.mean(torch.abs(x), dim=dims, keepdim=self.keep_dim)
        else:
            # default is mean
            # same as AdaptiveAvgPool
            x = torch.mean(x, dim=dims, keepdim=self.keep_dim)
        return x

    def forward(self, x: Tensor) -> Tensor:
        # @deepkyu: [fx tracing] Always x.dim() == 4
        # if x.dim() == 4:
        #     dims = [-2, -1]
        # elif x.dim() == 5:
        #     dims = [-3, -2, -1]
        # else:
        #     raise NotImplementedError("Currently 2D and 3D global pooling supported")

        return self._global_pool(x, dims=(-2, -1))

    # def profile_module(self, input: Tensor) -> Tuple[Tensor, float, float]:
    #     input = self.forward(input)
    #     return input, 0.0, 0.0

    # def __repr__(self):
    #     return "{}(type={})".format(self.__class__.__name__, self.pool_type)


class Focus(nn.Module):
    """Focus width and height information into channel space."""

    def __init__(self, in_channels, out_channels, ksize=1, stride=1, act_type="silu"):
        super().__init__()
        self.conv = ConvLayer(in_channels=in_channels * 4,
                              out_channels=out_channels,
                              kernel_size=ksize,
                              stride=stride,
                              act_type=act_type)

    def forward(self, x):
        # shape of x (b,c,w,h) -> y(b,4c,w/2,h/2)
        patch_top_left = x[..., ::2, ::2]
        patch_top_right = x[..., ::2, 1::2]
        patch_bot_left = x[..., 1::2, ::2]
        patch_bot_right = x[..., 1::2, 1::2]
        x = torch.cat(
            (
                patch_top_left,
                patch_bot_left,
                patch_top_right,
                patch_bot_right,
            ),
            dim=1,
        )
        return self.conv(x)


class CSPLayer(nn.Module):
    """
    C3 in yolov5, CSP Bottleneck with 3 convolutions

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        n (int): Number of Bottlenecks. Default: 1
        shortcut (bool): Whether to use shortcut connections. Default: True
        expansion (float): Channel expansion factor. Default: 0.5
        depthwise (bool): Whether to use depthwise separable convolutions. Default: False
        act_type (str): Activation function type. Default: "silu"
        layer_type (str): Type of CSP layer ("csp", "csprep", or "repncsp"). Default: "csp"
    """


    def __init__(
        self,
        in_channels,
        out_channels,
        n=1,
        shortcut=True,
        expansion=0.5,
        depthwise=False,
        act_type="silu",
        layer_type: Optional[str] = "csp",
        **kwargs
    ):
        """
        Args:
            in_channels (int): input channels.
            out_channels (int): output channels.
            n (int): number of Bottlenecks. Default value: 1.
        """
        # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        VALID_LAYER_TYPE = ["csp", "csprep", "repncsp"]
        assert layer_type.lower() in VALID_LAYER_TYPE, f"Invalid layer_type: '{layer_type}'. Must be one of {VALID_LAYER_TYPE}"
        hidden_channels = int(out_channels * expansion)  # hidden channels
        self.conv1 = ConvLayer(in_channels=in_channels,
                               out_channels=hidden_channels,
                               kernel_size=1,
                               stride=1, act_type=act_type)
        self.conv2 = ConvLayer(in_channels=in_channels,
                              out_channels=hidden_channels,
                              kernel_size=1,
                              stride=1, act_type=act_type)

        block_mapping = {
            "csp": (DarknetBlock, True),
            "csprep": (RepVGGBlock, False),
            "repncsp": (RepNBottleneck, True)
        }

        block, self.concat = block_mapping[layer_type.lower()]

        if self.concat:
            self.conv3 = ConvLayer(in_channels=2 * hidden_channels,
                               out_channels=out_channels,
                               kernel_size=1,
                               stride=1, act_type=act_type)
        else:
            self.conv3 = ConvLayer(in_channels=hidden_channels,
                               out_channels=out_channels,
                               kernel_size=1,
                               stride=1, act_type=act_type)

        module_list = [
            block(
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                shortcut=shortcut,
                expansion=1.0,
                depthwise=depthwise,
                act_type=act_type,
                **kwargs
            )
            for _ in range(n)
        ]
        self.m = nn.Sequential(*module_list)

    def forward(self, x):
        x_1 = self.conv1(x)
        x_2 = self.conv2(x)
        x_1 = self.m(x_1)
        x = torch.cat((x_1, x_2), dim=1) if self.concat else x_1 + x_2
        return self.conv3(x)


class CSPRepLayer(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 num_blocks: int=3,
                 expansion: float=1.0,
                 bias: bool= False,
                 use_identity: Optional[bool]=True,
                 act: str="silu"):
        super(CSPRepLayer, self).__init__()
        warnings.warn(
            "CSPRepLayer is deprecated and will be removed in a future version. "
            "Please use CSPLayer with appropriate configuration instead.",
            DeprecationWarning,
            stacklevel=2
        )
        hidden_channels = int(out_channels * expansion)
        self.conv1 = ConvLayer(in_channels, hidden_channels, kernel_size=1, stride=1, bias=bias, act_type=act)
        self.conv2 = ConvLayer(in_channels, hidden_channels, kernel_size=1, stride=1, bias=bias, act_type=act)
        self.bottlenecks = nn.Sequential(*[
            RepVGGBlock(hidden_channels, hidden_channels, act_type=act, use_identity=use_identity) for _ in range(num_blocks)
        ])
        if hidden_channels != out_channels:
            self.conv3 = ConvLayer(hidden_channels, out_channels, kernel_size=1, stride=1, bias=bias, act_type=act)
        else:
            self.conv3 = nn.Identity()

    def forward(self, x):
        x_1 = self.conv1(x)
        x_1 = self.bottlenecks(x_1)
        x_2 = self.conv2(x)
        return self.conv3(x_1 + x_2)


class SPPBottleneck(nn.Module):
    """Spatial pyramid pooling layer used in YOLOv3-SPP"""

    def __init__(
        self, in_channels, out_channels, kernel_sizes=(5, 9, 13), act_type="silu"
    ):
        super().__init__()
        hidden_channels = in_channels // 2
        self.conv1 = ConvLayer(in_channels=in_channels, out_channels=hidden_channels,
                               kernel_size=1, stride=1, act_type=act_type)
        self.m = nn.ModuleList(
            [
                Pool2d(method='max', kernel_size=ks, stride=1, padding=ks // 2)
                for ks in kernel_sizes
            ]
        )
        conv2_channels = hidden_channels * (len(kernel_sizes) + 1)
        self.conv2 = ConvLayer(in_channels=conv2_channels, out_channels=out_channels,
                               kernel_size=1, stride=1, act_type=act_type)

    def forward(self, x):
        x = self.conv1(x)
        x = torch.cat([x] + [m(x) for m in self.m], dim=1)
        x = self.conv2(x)
        return x


class ShuffleV2Block(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int,
        kernel_size: int,
        stride: int,
    ):
        super().__init__()
        assert stride in [1, 2], "Stride must be either 1 or 2"

        self.stride = stride
        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.padding = kernel_size // 2

        self.out_channels = out_channels - in_channels

        self.branch_main = self._create_main_branch()
        self.branch_proj = self._create_proj_branch() if stride == 2 else None

    def _create_main_branch(self) -> nn.Sequential:
        return nn.Sequential(
            ConvLayer(self.in_channels, self.hidden_channels, 1, 1, padding=0),
            ConvLayer(self.hidden_channels, self.hidden_channels, self.kernel_size,
                      stride=self.stride, padding=self.padding, groups=self.hidden_channels, use_act=False),
            ConvLayer(self.hidden_channels, self.out_channels, 1, 1, padding=0),
        )

    def _create_proj_branch(self) -> nn.Sequential:
        return nn.Sequential(
            ConvLayer(self.in_channels, self.in_channels, self.kernel_size,
                      self.stride, padding=self.padding, groups=self.in_channels, use_act=False),
            ConvLayer(self.in_channels, self.in_channels, 1, 1, padding=0),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.stride == 1:
            x_proj, x = self.channel_shuffle(x)
            return torch.cat((x_proj, self.branch_main(x)), 1)
        else:
            return torch.cat((self.branch_proj(x), self.branch_main(x)), 1)

    def channel_shuffle(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        b, c, h, w = x.shape
        x = x.reshape(b * c // 2, 2, h * w)
        x = x.permute(1, 0, 2)
        x = x.reshape(2, -1, c // 2, h, w)
        return x[0], x[1]


# Newly defined because of slight difference with Bottleneck of custom.py
class DarknetBlock(nn.Module):
    # Standard bottleneck
    def __init__(
        self,
        in_channels,
        out_channels,
        shortcut=True,
        expansion=0.5,
        depthwise=False,
        act_type="silu",
    ):
        super().__init__()
        hidden_channels = int(out_channels * expansion)
        self.conv1 = ConvLayer(in_channels=in_channels, out_channels=hidden_channels,
                                kernel_size=1, stride=1, act_type=act_type)
        if depthwise:
            self.conv2 = SeparableConvLayer(in_channels=hidden_channels, out_channels=out_channels,
                                            kernel_size=3, stride=1, act_type=act_type)
        else:
            self.conv2 = ConvLayer(in_channels=hidden_channels, out_channels=out_channels,
                                    kernel_size=3, stride=1, act_type=act_type)
        self.use_add = shortcut and in_channels == out_channels

    def forward(self, x):
        y = self.conv2(self.conv1(x))
        if self.use_add:
            y = y + x
        return y


class LayerScale2d(nn.Module):
    """
        Based on timm implementation.
    """
    def __init__(self, dim: int, init_values: float = 1e-5, inplace: bool = False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        gamma = self.gamma.view(1, -1, 1, 1)
        return x.mul_(gamma) if self.inplace else x * gamma


class ELAN(nn.Module):
    """
    unified ELAN structure.
    It supports ['basic', 'repncsp'] ELAN structure.
    This implementation is based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py.
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 part_channels: int,
                 process_channels: Optional[int] = None,
                 act_type: Optional[str] = None,
                 layer_type: Optional[str] = None,
                 n: Optional[int] = 1,
                 **kwargs
                 ):
        super().__init__()
        VALID_LAYER_TYPE = ["basic", "repncsp"]
        if layer_type is None:
            layer_type = "basic"
        assert layer_type.lower() in VALID_LAYER_TYPE, f"Invalid layer_type: '{layer_type}'. Must be one of {VALID_LAYER_TYPE}"


        if process_channels is None:
            process_channels = part_channels // 2

        self.conv1 = ConvLayer(in_channels, part_channels, kernel_size=1, act_type=act_type)
        if layer_type.lower() == "basic":
            self.conv2 = ConvLayer(part_channels // 2, process_channels, kernel_size=3, act_type=act_type)
        elif layer_type.lower() == "repncsp":
            self.conv2 = nn.Sequential(
                CSPLayer(part_channels // 2, process_channels, layer_type="repncsp", act_type=act_type, n=n, **kwargs),
                ConvLayer(process_channels, process_channels, kernel_size=3, padding=1, act_type=act_type)
            )

        if layer_type.lower() == "basic":
            self.conv3 = ConvLayer(process_channels, process_channels, kernel_size=3, act_type=act_type)
        elif layer_type.lower() == "repncsp":
            self.conv3 = nn.Sequential(
                CSPLayer(process_channels, process_channels, layer_type="repncsp", act_type=act_type, n=n, **kwargs),
                ConvLayer(process_channels, process_channels, kernel_size=3, padding=1, act_type=act_type)
            )
        self.conv4 = ConvLayer(part_channels + 2 * process_channels, out_channels, kernel_size=1, act_type=act_type)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        x1, x2 = self.conv1(x).chunk(2, 1)
        x3 = self.conv2(x2)
        x4 = self.conv3(x3)
        x5 = self.conv4(torch.cat([x1, x2, x3, x4], dim=1))
        return x5

class AConv(nn.Module):
    """
        Based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py

        A module that combines average pooling and convolution operations.
        Performs average pooling followed by convolution with optional activation.

        Args:
            in_channels (int): Number of input channels
            out_channels (int): Number of output channels
            kernel_size (int, optional): Kernel size for convolution. Default: 3
            stride (int, optional): Stride for convolution. Default: 2
            pool_kernel_size (int, optional): Kernel size for average pooling. Default: 2
            pool_stride (int, optional): Stride for average pooling. Default: 1
            act_type (str, optional): Activation function type. Default: None
    """
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 2,
        pool_kernel_size: int = 2,
        pool_stride: int = 1,
        act_type: Optional[str] = None,
    ) -> None:
        """
        Args:
            x (Union[Tensor, Proxy]): Input tensor

        Returns:
            Union[Tensor, Proxy]: Output tensor after pooling and convolution
        """
        super().__init__()

        self.layers = nn.Sequential(
            nn.AvgPool2d(
                kernel_size=pool_kernel_size,
                stride=pool_stride
            ),
            ConvLayer(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                act_type=act_type
            )
        )

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        return self.layers(x)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self.layers}"


class ADown(nn.Module):
    """
        Based on https://github.com/WongKinYiu/YOLO/blob/b96c8eaec16cfcabbf79947d98d2c575f0a114ad/yolo/model/module.py.
        A module that combines average and max pooling with convolution operations for feature rediction.

        Args:
            in_channels (int): Number of input channels
            out_channels (int): NUmber of output channels
    """
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act_type: Optional[str]= 'silu',
        ) -> None:
        super().__init__()
        half_in_channels = in_channels // 2
        half_out_channels = out_channels // 2
        mid_layer = {"kernel_size": 3, "stride": 2}
        self.avg_pool = Pool2d(method="avg", kernel_size=2, stride=1)
        self.conv1 = ConvLayer(half_in_channels, half_out_channels, act_type=act_type, **mid_layer)
        self.max_pool = Pool2d(method="max", **mid_layer, padding=(1, 1))
        self.conv2 = ConvLayer(half_in_channels, half_out_channels, kernel_size=1, act_type=act_type)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        x = self.avg_pool(x)
        x1, x2 = x.chunk(2, dim=1)
        x1 = self.conv1(x1)
        x2 = self.max_pool(x2)
        x2 = self.conv2(x2)
        return torch.cat((x1, x2), dim=1)


class SPPCSPLayer(nn.Module):
    """
        Based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py
    """
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_sizes: Tuple[int] = (5, 9, 13),
        expansion: float = 0.5,
        act_type: str = "silu",

    ):
        super().__init__()
        hidden_channels = int(2 * out_channels * expansion)
        self.pre_conv = nn.Sequential(
            ConvLayer(in_channels, hidden_channels, kernel_size=1, act_type=act_type),
            ConvLayer(hidden_channels, hidden_channels, kernel_size=3, act_type=act_type),
            ConvLayer(hidden_channels, hidden_channels, kernel_size=1, act_type=act_type),
        )
        self.short_conv = ConvLayer(in_channels, hidden_channels, kernel_size=1, act_type=act_type)
        self.paddings = [auto_pad(kernel_size) for kernel_size in kernel_sizes]
        self.pools = nn.ModuleList([nn.MaxPool2d(kernel_size=kernel_size, stride=1, padding=padding) for kernel_size, padding in zip(kernel_sizes, self.paddings)])
        self.post_conv = nn.Sequential(*[ConvLayer(4 * hidden_channels, hidden_channels, kernel_size=1, act_type=act_type),
                                         ConvLayer(hidden_channels, hidden_channels, kernel_size=3, act_type=act_type)])
        self.merge_conv = ConvLayer(2 * hidden_channels, out_channels, kernel_size=1, act_type=act_type)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        features = [self.pre_conv(x)]
        for pool in self.pools:
            features.append(pool(features[-1]))
        features = torch.cat(features, dim=1)
        y1 = self.post_conv(features)
        y2 = self.short_conv(x)
        y = torch.cat((y1, y2), dim=1)
        return self.merge_conv(y)


class SPPELAN(nn.Module):
    """
    SPPELAN module cpmprising multiple pooling and convolution layers.
    """
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: Optional[int] = None,
        act_type: Optional[str] = "silu"
    ):
        super(SPPELAN, self).__init__()
        hidden_channels = hidden_channels or out_channels // 2

        self.conv1 = ConvLayer(in_channels, hidden_channels, kernel_size=1, act_type=act_type)
        self.pools = nn.ModuleList([nn.MaxPool2d(kernel_size=5, stride=1, padding=auto_pad(kernel_size=5)) for _ in range(3)])
        self.conv5 = ConvLayer(4 * hidden_channels, out_channels, kernel_size=1, act_type=act_type)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        features = [self.conv1(x)]
        for pool in self.pools:
            features.append(pool(features[-1]))
        return self.conv5(torch.cat(features, dim=1))


class Anchor2Vec(nn.Module):
    """
        This implementation is based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py.
    """
    def __init__(self,
                 reg_max: int=16):
        super().__init__()
        self.reg_max = reg_max
        self.num_predictions = 4 # Number of predictions per anchor
        reverse_reg = torch.arange(reg_max, dtype=torch.float32).view(1, reg_max, 1, 1, 1)
        self.anchor2vec = nn.Conv3d(in_channels=reg_max, out_channels=1, kernel_size=1, bias=False)
        self.anchor2vec.weight = nn.Parameter(reverse_reg, requires_grad=False)

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        """
        Args:
            x (Tensor): Input tensor of shape (batch_size, channels, height, width)

        Returns:
            Tuple[Tensor, Tensor]: Tuple of (anchor_tensor, vector_tensor)
            where anchor_tensor has shape (batch_size, r, 4, height, width)
            and vector_tensor has shape (batch_size, height, width)
        """
        batch_size, _, height, width = x.shape
        anchor_x = x.view(batch_size, self.num_predictions, self.reg_max, height, width)
        anchor_x = anchor_x.permute(0, 2, 1, 3, 4)
        self.anchor2vec = self.anchor2vec.to(x.device)
        vector_x = self.anchor2vec(anchor_x.softmax(dim=1))[:, 0]
        return anchor_x, vector_x



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from typing import Dict, List

import torch
import torch.nn as nn
import torch.nn.functional as F

from netspresso_trainer.utils.bbox_utils import transform_bbox

from .yolox import IOUloss, YOLOXLoss


class YOLOFastestLoss(YOLOXLoss):
    def __init__(self, anchors, l1_activate_epoch=None, cur_epoch=None, **kwargs) -> None:
        super().__init__(l1_activate_epoch, cur_epoch, **kwargs)
        self.iou_loss = IOUloss(reduction="none", loss_type="giou")
        self.anchors = [torch.tensor(anchor, dtype=torch.float).view(-1, 2) for anchor in anchors]
        self.num_anchors = self.anchors[0].size(0)

    def use_l1_update(self):
        return False

    def get_output_and_grid(self, output, k, stride, dtype):
        grid = self.grids[k]
        device = output.device
        batch_size = output.shape[0]
        hsize, wsize = output.shape[-2:]
        if grid.shape[2:4] != output.shape[2:4]:
            yv, xv = torch.meshgrid(torch.arange(hsize), torch.arange(wsize), indexing="ij")
            grid = torch.stack((xv, yv), 2).repeat(self.num_anchors,1,1,1).view(1, self.num_anchors, hsize, wsize, 2).type(dtype)
            self.grids[k] = grid
        anchors = self.anchors[k].view(1, self.num_anchors, 1, 1, 2).to(device)
        output = output.permute(0, 1, 3, 4, 2)
        output = torch.cat([
            (output[..., :2].sigmoid() + grid) * stride,
            2. * (torch.tanh(output[..., 2:4]/2 -.549306) + 1.) * anchors,
            output[..., 4:]
        ], dim=-1).reshape(
            batch_size, hsize * wsize * self.num_anchors, -1
        )
        return output, grid.view(1, -1, 2)

    def forward(self, out: List, target: Dict) -> torch.Tensor:
        self.use_l1 = self.use_l1_update()

        out = out['pred']
        x_shifts = []
        y_shifts = []
        expanded_strides = []

        self.grids = [torch.zeros(1)] * len(out)
        self.num_classes = target['num_classes']
        img_size = target['img_size']

        target = target['gt']

        out_for_loss = []
        for k, o in enumerate(out):
            stride_this_level = img_size // o.size(-1)

            o, grid = self.get_output_and_grid(
                o, k, stride_this_level, o.type()
            )
            x_shifts.append(grid[:, :, 0])
            y_shifts.append(grid[:, :, 1])
            expanded_strides.append(
                torch.zeros(1, grid.shape[1])
                .fill_(stride_this_level)
                .type_as(o)
            )
            out_for_loss.append(o)

        # YOLOX model learns box cxcywh format directly,
        # but our detection dataloader gives xyxy format.
        for i in range(len(target)):
            target[i]['boxes'] = transform_bbox(target[i]['boxes'], "xyxy -> cxcywh")

        # Ready for l1 loss
        origin_preds = []
        for o in out:
            out_for_l1 = o.view(o.shape[0], self.num_anchors, -1, o.shape[-2], o.shape[-1]).permute(0, 1, 3, 4, 2)
            reg_output = out_for_l1[..., :4]
            batch_size = reg_output.shape[0]
            reg_output = reg_output.reshape(
                batch_size, -1, 4
            )
            origin_preds.append(reg_output.clone())

        total_loss, iou_loss, conf_loss, cls_loss, l1_loss, num_fg = self.get_losses(
                    None,
                    x_shifts,
                    y_shifts,
                    expanded_strides,
                    target,
                    torch.cat(out_for_loss, 1),
                    origin_preds,
                    dtype=out[0].dtype,
                )

        # TODO: return as dict
        return total_loss



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

import numpy as np

from ..base import BaseMetric

IGNORE_INDEX_NONE_VALUE = 255


#TODO: Temporarily added IoUMeter. MetricMeter should be more generalized.
class IoUMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, num_classes, name: str, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.num_classes = num_classes
        self.reset()

    def reset(self):
        self.intersection = np.zeros(self.num_classes)
        self.union = np.zeros(self.num_classes)

    def update(self, intersection, union) -> None:
        self.intersection += intersection
        self.union += union

    def __str__(self):
        return f'{self.name} {np.nanmean(self.intersection / self.union):6.2f}'

    @property
    def avg(self) -> float:
        return np.nanmean(self.intersection / self.union)


class SegmentationMetricAdaptor:
    '''
        Adapter to process redundant operations for the metrics.
    '''
    def __init__(self, metric_names) -> None:
        self.metric_names = metric_names

    def __call__(self, predictions: List[dict], targets: List[dict]):
        return {} # Do nothing


# TODO: Unify repeated code
class mIoU(BaseMetric):
    def __init__(self, num_classes, classwise_analysis, ignore_index=IGNORE_INDEX_NONE_VALUE, **kwargs):
        metric_name = 'mIoU' # Name for logging
        super().__init__(metric_name=metric_name, num_classes=num_classes, classwise_analysis=classwise_analysis)
        self.ignore_index = ignore_index if ignore_index is not None else IGNORE_INDEX_NONE_VALUE
        self.metric_meter = IoUMeter(num_classes=self.num_classes, name='iou') #TODO: Temporarily added IoUMeter.

    def intersection_and_union(self, output, target):

        # 'K' classes, output and target sizes are N or N * L or N * H * W, each value in range 0 to K - 1.
        assert (len(output.shape) in [1, 2, 3])

        assert output.shape == target.shape
        output = output.reshape(-1)
        target = target.reshape(-1)
        output[target == self.ignore_index] = self.ignore_index
        intersection = output[output == target]
        #area_intersection = torch.histc(intersection, bins=self.K, min=0, max=self.K-1)
        #area_output = torch.histc(output, bins=self.K, min=0, max=self.K-1)
        #area_target = torch.histc(target, bins=self.K, min=0, max=self.K-1)
        area_intersection = np.histogram(intersection, bins=np.linspace(0, self.num_classes, self.num_classes+1))[0]
        area_output = np.histogram(output, bins=np.linspace(0, self.num_classes, self.num_classes+1))[0]
        area_target = np.histogram(target, bins=np.linspace(0, self.num_classes, self.num_classes+1))[0]
        area_union = area_output + area_target - area_intersection

        intersection, union, target, output = area_intersection, area_union, area_target, area_output

        return {
            'intersection': intersection,
            'union': union,
            'target': target,
            'output': output
        }

    def calibrate(self, pred, target, **kwargs):
        metrics = self.intersection_and_union(pred, target)

        if self.classwise_analysis: # TODO: Compute in a better way
            for cls_meter, cls_intersection, cls_union in zip(self.classwise_metric_meters, metrics['intersection'], metrics['union']):
                if cls_union != 0:
                    cls_meter.update(cls_intersection, cls_union)

        self.metric_meter.update(metrics['intersection'], metrics['union'])


class PixelAccuracy(BaseMetric):
    def __init__(self, num_classes, classwise_analysis, ignore_index=IGNORE_INDEX_NONE_VALUE, **kwargs):
        metric_name = 'Pixel_acc' # Name for logging
        super().__init__(metric_name=metric_name, num_classes=num_classes, classwise_analysis=classwise_analysis)
        self.ignore_index = ignore_index if ignore_index is not None else IGNORE_INDEX_NONE_VALUE

    def intersection_and_union(self, output, target):

        # 'K' classes, output and target sizes are N or N * L or N * H * W, each value in range 0 to K - 1.
        assert (len(output.shape) in [1, 2, 3])

        assert output.shape == target.shape
        output = output.reshape(-1)
        target = target.reshape(-1)
        output[target == self.ignore_index] = self.ignore_index
        intersection = output[output == target]

        area_intersection = np.histogram(intersection, bins=np.linspace(0, self.num_classes, self.num_classes+1))[0]
        area_target = np.histogram(target, bins=np.linspace(0, self.num_classes, self.num_classes+1))[0]

        intersection, target = area_intersection, area_target

        return {
            'intersection': intersection,
            'target': target,
        }

    def calibrate(self, pred, target, **kwargs):
        B = pred.shape[0]

        metrics = self.intersection_and_union(pred, target)

        if self.classwise_analysis: # TODO: Compute in a better way
            for cls_meter, cls_intersection, cls_target in zip(self.classwise_metric_meters, metrics['intersection'], metrics['target']):
                if cls_target != 0:
                    cls_meter.update(cls_intersection, cls_target)

        self.metric_meter.update(sum(metrics['intersection']) / (sum(metrics['target']) + 1e-10), n=B)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the Darknet implementation of Megvii.
https://github.com/Megvii-BaseDetection/YOLOX/blob/main/yolox/models/darknet.py
"""
from typing import Dict, Optional, List

from omegaconf import DictConfig
import torch
from torch import nn

from ...op.custom import ConvLayer, CSPLayer, Focus, SPPBottleneck, SeparableConvLayer
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['cspdarknet']
SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']


class CSPDarknet(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'CSPDarknet is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        super().__init__()

        out_features=("dark3", "dark4", "dark5")
        assert out_features, "please provide output features of Darknet"

        dep_mul = params.dep_mul
        wid_mul = params.wid_mul
        act_type = params.act_type
        depthwise = params.depthwise
        
        self.out_features = out_features
        Conv = SeparableConvLayer if depthwise else ConvLayer

        base_channels = int(wid_mul * 64)  # 64
        base_depth = max(round(dep_mul * 3), 1)  # 3

        # stem
        self.stem = Focus(3, base_channels, ksize=3, act_type=act_type)

        # dark2
        self.dark2 = nn.Sequential(
            Conv(in_channels=base_channels, 
                 out_channels=base_channels * 2, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 2,
                base_channels * 2,
                n=base_depth,
                depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark3
        self.dark3 = nn.Sequential(
            Conv(in_channels=base_channels * 2, 
                 out_channels=base_channels * 4, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 4,
                base_channels * 4,
                n=base_depth * 3,
                depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark4
        self.dark4 = nn.Sequential(
            Conv(in_channels=base_channels * 4,
                 out_channels=base_channels * 8, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 8,
                base_channels * 8,
                n=base_depth * 3,
                depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark5
        self.dark5 = nn.Sequential(
            Conv(in_channels=base_channels * 8, 
                 out_channels=base_channels * 16, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            SPPBottleneck(base_channels * 16, base_channels * 16, act_type=act_type),
            CSPLayer(
                base_channels * 16,
                base_channels * 16,
                n=base_depth,
                shortcut=False,
                depthwise=depthwise,
                act_type=act_type,
            ),
        )

        self.avgpool = nn.AdaptiveAvgPool2d(1)

        predefined_out_features = {'dark2': base_channels * 2, 'dark3': base_channels * 4, 
                                   'dark4': base_channels * 8, 'dark5': base_channels * 16}
        self._feature_dim = predefined_out_features['dark5']
        self._intermediate_features_dim = [predefined_out_features[out_feature] for out_feature in out_features]

        # Initialize
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)

    def forward(self, x):
        outputs_dict = {}
        x = self.stem(x)
        outputs_dict["stem"] = x
        x = self.dark2(x)
        outputs_dict["dark2"] = x
        x = self.dark3(x)
        outputs_dict["dark3"] = x
        x = self.dark4(x)
        outputs_dict["dark4"] = x
        x = self.dark5(x)
        outputs_dict["dark5"] = x

        if self.use_intermediate_features:
            all_hidden_states = [outputs_dict[out_name] for out_name in self.out_features]
            return BackboneOutput(intermediate_features=all_hidden_states)
        
        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)
    
    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def cspdarknet(task, conf_model_backbone) -> CSPDarknet:
    return CSPDarknet(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import warnings

import torch
from loguru import logger
from torch.optim.lr_scheduler import _LRScheduler


class PolynomialLRWithWarmUp(_LRScheduler):
    """Decays the learning rate of each parameter group using a polynomial function
    in the given total_iters with warmup When last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        warmup_iters (int): The number of steps that the scheduler finishes to warmup the learning rate.
        total_iters (int): The number of steps that the scheduler decays the learning rate.
        power (int): The power of the polynomial.
    """
    def __init__(
        self,
        optimizer,
        scheduler_conf,
        training_epochs,
    ):
        self.warmup_iters = scheduler_conf.warmup_epochs
        self.total_iters = scheduler_conf.end_epoch
        self.warmup_bias_lr = scheduler_conf.warmup_bias_lr
        self.min_lr = scheduler_conf.min_lr
        self.power = scheduler_conf.power
        super().__init__(optimizer)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)

        if self.last_epoch > self.total_iters:
            return [group["lr"] for group in self.optimizer.param_groups]

        if self.last_epoch >= 0 and self.last_epoch < self.warmup_iters:
            return [self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr)
                    for base_lr in self.base_lrs]

        decay_steps = self.total_iters - self.warmup_iters
        decay_current_step = self.last_epoch - self.warmup_iters
        decay_factor = ((1.0 - decay_current_step / decay_steps) / (1.0 - (decay_current_step - 1) / decay_steps)) ** self.power
        return [self.min_lr + (group["lr"] - self.min_lr) * decay_factor for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        decay_steps = self.total_iters - self.warmup_iters
        return [
            (
                min(
                    self.warmup_bias_lr + (base_lr - self.warmup_bias_lr) * float(self.last_epoch + 1) / float(max(1, self.warmup_iters)),
                    self.min_lr + (base_lr - self.min_lr) * (1.0 - min(self.last_epoch - self.warmup_iters, decay_steps) / decay_steps) ** self.power
                )
            )
            for base_lr in self.base_lrs
        ]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the SimCCLabel decode implementation of mmpose.
https://github.com/open-mmlab/mmpose
"""
from typing import Optional

import torch

from ..models.utils import ModelOutput


class PoseEstimationPostprocessor():
    def __init__(self, conf_model):
        params = conf_model.postprocessor.params
        self.simcc_split_ratio = params.simcc_split_ratio

    def get_simcc_maximum(self, simcc_x, simcc_y, apply_softmax: bool = False):
        assert simcc_x.ndim == 2 or simcc_x.ndim == 3, (f'Invalid shape {simcc_x.shape}')
        assert simcc_y.ndim == 2 or simcc_y.ndim == 3, (f'Invalid shape {simcc_y.shape}')
        assert simcc_x.ndim == simcc_y.ndim, (f'{simcc_x.shape} != {simcc_y.shape}')

        N, K, _ = simcc_x.shape
        simcc_x = simcc_x.reshape(N * K, -1)
        simcc_y = simcc_y.reshape(N * K, -1)

        # if apply_softmax:
        #     simcc_x = simcc_x - np.max(simcc_x, axis=1, keepdims=True)
        #     simcc_y = simcc_y - np.max(simcc_y, axis=1, keepdims=True)
        #     ex, ey = np.exp(simcc_x), np.exp(simcc_y)
        #     simcc_x = ex / np.sum(ex, axis=1, keepdims=True)
        #     simcc_y = ey / np.sum(ey, axis=1, keepdims=True)

        x_locs = torch.argmax(simcc_x, axis=1)
        y_locs = torch.argmax(simcc_y, axis=1)
        locs = torch.stack((x_locs, y_locs), axis=-1).to(torch.float32)
        max_val_x = torch.amax(simcc_x, axis=1)
        max_val_y = torch.amax(simcc_y, axis=1)

        mask = max_val_x > max_val_y
        max_val_x[mask] = max_val_y[mask]
        vals = max_val_x
        locs[vals <= 0.] = -1

        if N:
            locs = locs.reshape(N, K, 2)
            vals = vals.reshape(N, K)

        return locs, vals

    def __call__(self, outputs: ModelOutput):
        pred = outputs['pred']
        coord_split = pred.shape[-1] // 2

        simcc_x = pred[..., :coord_split]
        simcc_y = pred[..., coord_split:]

        keypoints, scores = self.get_simcc_maximum(simcc_x, simcc_y)

        # TODO: use_dark option
        # if self.use_dark:
        #     x_blur = int((self.sigma[0] * 20 - 7) // 3)
        #     y_blur = int((self.sigma[1] * 20 - 7) // 3)
        #     x_blur -= int((x_blur % 2) == 0)
        #     y_blur -= int((y_blur % 2) == 0)
        #     keypoints[:, :, 0] = refine_simcc_dark(keypoints[:, :, 0], simcc_x,
        #                                         x_blur)
        #     keypoints[:, :, 1] = refine_simcc_dark(keypoints[:, :, 1], simcc_y,
        #                                         y_blur)

        keypoints /= self.simcc_split_ratio

        # TODO: decode_visibility option
        # if self.decode_visibility:
        #     _, visibility = get_simcc_maximum(
        #         simcc_x * self.decode_beta * self.sigma[0],
        #         simcc_y * self.decode_beta * self.sigma[1],
        #         apply_softmax=True)
        #     return keypoints, (scores, visibility)
        # else:
        #     return keypoints, scores
        return keypoints.detach().cpu().numpy() #, scores.detach().cpu().numpy()



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from typing import Any, Dict, List, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn import BCEWithLogitsLoss

from netspresso_trainer.models.op.custom import Anchor2Vec
from netspresso_trainer.utils.bbox_utils import generate_anchors

from .yolox import IOUloss

"""
    This implementation is based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/tools/loss_functions.py.
"""


def calculate_iou(bbox1, bbox2, metrics="iou") -> Tensor:
    #TODO: It should be consolidated into bbox_utils
    metrics = metrics.lower()
    EPS = 1e-7
    dtype = bbox1.dtype
    bbox1 = bbox1.to(torch.float32)
    bbox2 = bbox2.to(torch.float32)

    # Expand dimensions if necessary
    if bbox1.ndim == 2 and bbox2.ndim == 2:
        bbox1 = bbox1.unsqueeze(1)  # (Ax4) -> (Ax1x4)
        bbox2 = bbox2.unsqueeze(0)  # (Bx4) -> (1xBx4)
    elif bbox1.ndim == 3 and bbox2.ndim == 3:
        bbox1 = bbox1.unsqueeze(2)  # (BZxAx4) -> (BZxAx1x4)
        bbox2 = bbox2.unsqueeze(1)  # (BZxBx4) -> (BZx1xBx4)

    # Calculate intersection coordinates
    xmin_inter = torch.max(bbox1[..., 0], bbox2[..., 0])
    ymin_inter = torch.max(bbox1[..., 1], bbox2[..., 1])
    xmax_inter = torch.min(bbox1[..., 2], bbox2[..., 2])
    ymax_inter = torch.min(bbox1[..., 3], bbox2[..., 3])

    # Calculate intersection area
    intersection_area = torch.clamp(xmax_inter - xmin_inter, min=0) * torch.clamp(ymax_inter - ymin_inter, min=0)

    # Calculate area of each bbox
    area_bbox1 = (bbox1[..., 2] - bbox1[..., 0]) * (bbox1[..., 3] - bbox1[..., 1])
    area_bbox2 = (bbox2[..., 2] - bbox2[..., 0]) * (bbox2[..., 3] - bbox2[..., 1])

    # Calculate union area
    union_area = area_bbox1 + area_bbox2 - intersection_area

    # Calculate IoU
    iou = intersection_area / (union_area + EPS)
    if metrics == "iou":
        return iou.to(dtype)

    # Calculate centroid distance
    cx1 = (bbox1[..., 2] + bbox1[..., 0]) / 2
    cy1 = (bbox1[..., 3] + bbox1[..., 1]) / 2
    cx2 = (bbox2[..., 2] + bbox2[..., 0]) / 2
    cy2 = (bbox2[..., 3] + bbox2[..., 1]) / 2
    cent_dis = (cx1 - cx2) ** 2 + (cy1 - cy2) ** 2

    # Calculate diagonal length of the smallest enclosing box
    c_x = torch.max(bbox1[..., 2], bbox2[..., 2]) - torch.min(bbox1[..., 0], bbox2[..., 0])
    c_y = torch.max(bbox1[..., 3], bbox2[..., 3]) - torch.min(bbox1[..., 1], bbox2[..., 1])
    diag_dis = c_x**2 + c_y**2 + EPS

    diou = iou - (cent_dis / diag_dis)
    if metrics == "diou":
        return diou.to(dtype)

    # Compute aspect ratio penalty term
    arctan = torch.atan((bbox1[..., 2] - bbox1[..., 0]) / (bbox1[..., 3] - bbox1[..., 1] + EPS)) - torch.atan(
        (bbox2[..., 2] - bbox2[..., 0]) / (bbox2[..., 3] - bbox2[..., 1] + EPS)
    )
    v = (4 / (math.pi**2)) * (arctan**2)
    with torch.no_grad():
        alpha = v / (v - iou + 1 + EPS)
    # Compute CIoU
    ciou = diou - alpha * v
    return ciou.to(dtype)



class BoxMatcher:
    # TODO: It should be consolidated into bbox_utils
    def __init__(self,
                 class_num: int,
                 anchors: Tensor,

                 ) -> None:
        self.class_num = class_num
        self.anchors = anchors
        self.iou = "ciou"
        self.topk = 10 # TODO: It should be parameterized.
        self.factor = {"iou": 6.0, "cls": 0.5} # TODO: It should be parameterized.

    def get_valid_matrix(self, target_bbox: Tensor):
        """
        Get a boolean mask that indicates whether each target bounding box overlaps with each anchor.

        Args:
            target_bbox [batch x targets x 4]: The bounding box of each targets.
        Returns:
            [batch x targets x anchors]: A boolean tensor indicates if target bounding box overlaps with anchors.
        """
        Xmin, Ymin, Xmax, Ymax = target_bbox[:, :, None].unbind(3)
        anchors = self.anchors[None, None]  # add a axis at first, second dimension
        anchors_x, anchors_y = anchors.unbind(dim=3)
        target_in_x = (Xmin < anchors_x) & (anchors_x < Xmax)
        target_in_y = (Ymin < anchors_y) & (anchors_y < Ymax)
        target_on_anchor = target_in_x & target_in_y
        return target_on_anchor

    def get_cls_matrix(self, predict_cls: Tensor, target_cls: Tensor) -> Tensor:
        """
        Get the (predicted class' probabilities) corresponding to the target classes across all anchors

        Args:
            predict_cls [batch x anchors x class]: The predicted probabilities for each class across each anchor.
            target_cls [batch x targets]: The class index for each target.

        Returns:
            [batch x targets x anchors]: The probabilities from `pred_cls` corresponding to the class indices specified in `target_cls`.
        """
        predict_cls = predict_cls.transpose(1, 2)
        target_cls = target_cls.expand(-1, -1, predict_cls.size(2))
        cls_probabilities = torch.gather(predict_cls, 1, target_cls)
        return cls_probabilities

    def get_iou_matrix(self, predict_bbox, target_bbox) -> Tensor:
        """
        Get the IoU between each target bounding box and each predicted bounding box.

        Args:
            predict_bbox [batch x predicts x 4]: Bounding box with [x1, y1, x2, y2].
            target_bbox [batch x targets x 4]: Bounding box with [x1, y1, x2, y2].
        Returns:
            [batch x targets x predicts]: The IoU scores between each target and predicted.
        """
        return calculate_iou(target_bbox, predict_bbox, self.iou).clamp(0, 1)

    def filter_topk(self, target_matrix: Tensor, topk: int = 10) -> Tuple[Tensor, Tensor]:
        """
        Filter the top-k suitability of targets for each anchor.

        Args:
            target_matrix [batch x targets x anchors]: The suitability for each targets-anchors
            topk (int, optional): Number of top scores to retain per anchor.

        Returns:
            topk_targets [batch x targets x anchors]: Only leave the topk targets for each anchor
            topk_masks [batch x targets x anchors]: A boolean mask indicating the top-k scores' positions.
        """
        values, indices = target_matrix.topk(topk, dim=-1)
        topk_targets = torch.zeros_like(target_matrix, device=target_matrix.device)
        topk_targets.scatter_(dim=-1, index=indices, src=values)
        topk_masks = topk_targets > 0
        return topk_targets, topk_masks

    def filter_duplicates(self, target_matrix: Tensor, topk_mask: Tensor):
        """
        Filter the maximum suitability target index of each anchor.

        Args:
            target_matrix [batch x targets x anchors]: The suitability for each targets-anchors

        Returns:
            unique_indices [batch x anchors x 1]: The index of the best targets for each anchors
        """
        duplicates = (topk_mask.sum(1, keepdim=True) > 1).repeat([1, topk_mask.size(1), 1])
        max_idx = F.one_hot(target_matrix.argmax(1), topk_mask.size(1)).permute(0, 2, 1)
        topk_mask = torch.where(duplicates, max_idx, topk_mask)
        unique_indices = topk_mask.argmax(dim=1)
        return unique_indices[..., None], topk_mask.sum(1), topk_mask

    def __call__(self, target: Tensor, predict: Tuple[Tensor]) -> Tuple[Tensor, Tensor]:
        """Matches each target to the most suitable anchor.
        1. For each anchor prediction, find the highest suitability targets.
        2. Match target to the best anchor.
        3. Noramlize the class probilities of targets.

        Args:
            target: The ground truth class and bounding box information
                as tensor of size [batch x targets x 5].
            predict: Tuple of predicted class and bounding box tensors.
                Class tensor is of size [batch x anchors x class]
                Bounding box tensor is of size [batch x anchors x 4].

        Returns:
            anchor_matched_targets: Tensor of size [batch x anchors x (class + 4)].
                A tensor assigning each target/gt to the best fitting anchor.
                The class probabilities are normalized.
            valid_mask: Bool tensor of shape [batch x anchors].
                True if a anchor has a target/gt assigned to it.
        """
        predict_cls, predict_bbox = predict

        # return if target has no gt information.
        n_targets = target.shape[1]
        if n_targets == 0:
            device = predict_bbox.device
            align_cls = torch.zeros_like(predict_cls, device=device)
            align_bbox = torch.zeros_like(predict_bbox, device=device)
            valid_mask = torch.zeros(predict_cls.shape[:2], dtype=bool, device=device)
            anchor_matched_targets = torch.cat([align_cls, align_bbox], dim=-1)
            return anchor_matched_targets, valid_mask

        target_cls, target_bbox = target.split([1, 4], dim=-1)  # B x N x (C B) -> B x N x C, B x N x B
        target_cls = target_cls.long().clamp(0)

        # get valid matrix (each gt appear in which anchor grid)
        grid_mask = self.get_valid_matrix(target_bbox)

        # get iou matrix (iou with each gt bbox and each predict anchor)
        iou_mat = self.get_iou_matrix(predict_bbox, target_bbox)

        # get cls matrix (cls prob with each gt class and each predict class)
        cls_mat = self.get_cls_matrix(predict_cls.sigmoid(), target_cls)

        target_matrix = grid_mask * (iou_mat ** self.factor["iou"]) * (cls_mat ** self.factor["cls"])

        # choose topk
        topk_targets, topk_mask = self.filter_topk(target_matrix, topk=self.topk)

        # delete one anchor pred assign to mutliple gts
        unique_indices, valid_mask, topk_mask = self.filter_duplicates(iou_mat, topk_mask)

        align_bbox = torch.gather(target_bbox, 1, unique_indices.repeat(1, 1, 4))
        align_cls = torch.gather(target_cls, 1, unique_indices).squeeze(-1)
        align_cls = F.one_hot(align_cls, self.class_num)

        # normalize class ditribution
        iou_mat *= topk_mask
        target_matrix *= topk_mask
        max_target = target_matrix.amax(dim=-1, keepdim=True)
        max_iou = iou_mat.amax(dim=-1, keepdim=True)
        normalize_term = (target_matrix / (max_target + 1e-9)) * max_iou
        normalize_term = normalize_term.permute(0, 2, 1).gather(2, unique_indices)
        align_cls = align_cls * normalize_term * valid_mask[:, :, None]
        anchor_matched_targets = torch.cat([align_cls, align_bbox], dim=-1)
        return anchor_matched_targets, valid_mask.bool()




class BCELoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        # TODO: Refactor the device, should be assign by config
        # TODO: origin v9 using pos_weight == 1?
        self.bce = BCEWithLogitsLoss(reduction="none")

    def forward(self, predicts_cls: Tensor, targets_cls: Tensor, cls_norm: Tensor) -> Any:
        return self.bce(predicts_cls, targets_cls).sum() / cls_norm


class BoxLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(
        self, predicts_bbox: Tensor, targets_bbox: Tensor, valid_masks: Tensor, box_norm: Tensor, cls_norm: Tensor
    ) -> Any:
        valid_bbox = valid_masks[..., None].expand(-1, -1, 4)
        picked_predict = predicts_bbox[valid_bbox].view(-1, 4)
        picked_targets = targets_bbox[valid_bbox].view(-1, 4)

        iou = calculate_iou(picked_predict, picked_targets, "ciou").diag()
        loss_iou = 1.0 - iou
        loss_iou = (loss_iou * box_norm).sum() / cls_norm
        return loss_iou


class DFLoss(nn.Module):
    def __init__(self, anchor_grid, scaler, reg_max: int) -> None:
        super().__init__()
        self.anchors_norm = (anchor_grid / scaler[:, None])[None]
        self.reg_max = reg_max

    def forward(
        self, predicts_anc: Tensor, targets_bbox: Tensor, valid_masks: Tensor, box_norm: Tensor, cls_norm: Tensor
    ) -> Any:
        valid_bbox = valid_masks[..., None].expand(-1, -1, 4)
        bbox_lt, bbox_rb = targets_bbox.chunk(2, -1)
        targets_dist = torch.cat(((self.anchors_norm - bbox_lt), (bbox_rb - self.anchors_norm)), -1).clamp(
            0, self.reg_max - 1.01
        )
        picked_targets = targets_dist[valid_bbox].view(-1)
        picked_predict = predicts_anc[valid_bbox].view(-1, self.reg_max)

        label_left, label_right = picked_targets.floor(), picked_targets.floor() + 1
        weight_left, weight_right = label_right - picked_targets, picked_targets - label_left

        loss_left = F.cross_entropy(picked_predict, label_left.to(torch.long), reduction="none")
        loss_right = F.cross_entropy(picked_predict, label_right.to(torch.long), reduction="none")
        loss_dfl = loss_left * weight_left + loss_right * weight_right
        loss_dfl = loss_dfl.view(-1, 4).mean(-1)
        loss_dfl = (loss_dfl * box_norm).sum() / cls_norm
        return loss_dfl


class YOLOv9Loss(nn.Module):
    def __init__(self, l1_activate_epoch, cur_epoch=None, **kwargs):
        super().__init__()
        self.iou_loss = IOUloss(reduction="none", loss_type="ciou")
        self.bcewithlog_loss = nn.BCEWithLogitsLoss(reduction="none")
        self.anchor_grid = None
        self.scaler = None
        self.cls = BCELoss()
        self.iou = BoxLoss()
        self.reg_max = kwargs.get("reg_max", 16)
        self.anc2vec = Anchor2Vec(self.reg_max)
        self.aux_rate = 0.25 # TODO: should be controlled by config

    def get_output(self, output, anchor_grid, scaler):
        pred_bbox_reg, pred_bbox_anchor, pred_class_logits = [], [], []
        for layer_output in output:
            reg, class_logits = torch.split(layer_output, [layer_output.shape[1] - self.num_classes, self.num_classes], dim=1)
            bbox_anchor, bbox_reg = self.anc2vec(reg)
            b, c, _, _ = bbox_reg.shape
            reg = bbox_reg.view(b, c, -1).permute(0, 2, 1)
            pred_bbox_reg.append(reg)

            b, a, r, _, _ = bbox_anchor.shape
            anchor = bbox_anchor.view(b, a, r, -1).permute(0, 3, 2, 1)
            pred_bbox_anchor.append(anchor)

            b, c, _, _ = class_logits.shape
            logits = class_logits.view(b, c, -1).permute(0, 2, 1)
            pred_class_logits.append(logits)

        pred_bbox_reg = torch.concat(pred_bbox_reg, dim=1)
        pred_bbox_anchor = torch.concat(pred_bbox_anchor, dim=1)
        pred_class_logits = torch.concat(pred_class_logits, dim=1)

        pred_xyxy = pred_bbox_reg * scaler.view(1, -1, 1)
        lt, rb = pred_xyxy.chunk(2, dim=-1)
        pred_bbox_reg = torch.cat([anchor_grid - lt, anchor_grid + rb], dim=-1)
        return pred_class_logits, pred_bbox_anchor, pred_bbox_reg

    def forward(self, out: Union[List, Dict], target: Dict) -> Tensor:
        # aux_out = out['pred']['aux_outputs'] if isinstance(out['pred'], Dict) else None
        out = out['pred']['outputs'] if isinstance(out['pred'], Dict) else out['pred']
        device = out[0][0].device
        self.num_classes = target['num_classes']
        img_size = target['img_size']
        if isinstance(img_size, int):
            img_size = (img_size, img_size)

        target = target['gt']

        strides = []
        for _k, o in enumerate(out):
            stride_this_level = img_size[-1] // o.size(-1)
            strides.append(stride_this_level)
        anchor_grid, scaler = generate_anchors(img_size, strides)
        anchor_grid = anchor_grid.to(device)
        scaler = scaler.to(device)

        dfl = DFLoss(anchor_grid, scaler, self.reg_max)
        preds_cls, preds_anc, preds_box = self.get_output(out, anchor_grid=anchor_grid, scaler=scaler)
        # if aux_out:
        #     aux_preds_cls, aux_preds_anc, aux_preds_box = self.get_output(aux_out, anchor_grid=anchor_grid, scaler=scaler)
        # else:
        #     aux_preds_cls, aux_preds_anc, aux_preds_box = None, None, None

        matcher = BoxMatcher(self.num_classes, anchor_grid)
        max_samples = max([len(t['labels']) for t in target])
        labels = torch.zeros(len(target), max_samples, 5).to(device)
        labels[:, :, 0] = -1
        for batch_idx, t in enumerate(target):
            batch_boxes = t['boxes']
            batch_labels = t['labels']
            sample_num = len(batch_labels)
            labels[batch_idx, :sample_num, 1:5] = batch_boxes
            labels[batch_idx, :sample_num, 0] = batch_labels

        align_targets, valid_masks = matcher(labels, (preds_cls.detach(), preds_box.detach()))
        targets_cls, targets_bbox = self.separate_anchor(align_targets, scaler)
        preds_box = preds_box / scaler[None, :, None]
        cls_norm = max(targets_cls.sum(), 1)
        box_norm = targets_cls.sum(-1)[valid_masks]
        # if aux_out:
        #     aux_align_targets, aux_valid_masks = matcher(labels, (aux_preds_cls.detach(), aux_preds_box.detach()))
        #     aux_preds_box = aux_preds_box / scaler[None, :, None]
        #     aux_targets_cls, aux_targets_bbox = self.separate_anchor(aux_align_targets, scaler)
        #     aux_cls_norm = max(aux_targets_cls.sum(), 1)
        #     aux_box_norm = aux_targets_cls.sum(-1)[aux_valid_masks]
        #     ## -- CLS -- ##
        #     aux_loss_cls = self.cls(aux_preds_cls, aux_targets_cls, aux_cls_norm)
        #     ## -- IOU -- ##
        #     aux_loss_iou = self.iou(aux_preds_box, aux_targets_bbox, aux_valid_masks, aux_box_norm, aux_cls_norm)
        #     ## -- DFL -- ##
        #     aux_loss_dfl = dfl(aux_preds_anc, aux_targets_bbox, aux_valid_masks, aux_box_norm, aux_cls_norm)


        ## -- CLS -- ##
        loss_cls = self.cls(preds_cls, targets_cls, cls_norm)
        ## -- IOU -- ##
        loss_iou = self.iou(preds_box, targets_bbox, valid_masks, box_norm, cls_norm)
        ## -- DFL -- ##
        loss_dfl = dfl(preds_anc, targets_bbox, valid_masks, box_norm, cls_norm)

        # TODO: loss weights should be controlled by config
        # if aux_out:
        #     total_loss = 0.5 * (self.aux_rate * aux_loss_cls + loss_cls) + 1.5 * (self.aux_rate * aux_loss_dfl + loss_dfl) + 7.5 * (self.aux_rate * aux_loss_iou + loss_iou)
        # else:
        total_loss = 0.5 * loss_cls + 1.5 * loss_dfl + 7.5 * loss_iou
        return total_loss

    def separate_anchor(self, anchors, scaler):
        """
        separate anchor and bbouding box
        """
        anchors_cls, anchors_box = torch.split(anchors, (self.num_classes, 4), dim=-1)
        anchors_box = anchors_box / scaler[None, :, None]
        return anchors_cls, anchors_box



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the Torchvision implementation of MobileNetV3.
https://pytorch.org/vision/main/_modules/torchvision/models/mobilenetv3.html
"""
from typing import List, Dict, Optional

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import ConvLayer, InvertedResidual
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['mobilenetv3']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']


class MobileNetV3(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'MobileNetV3 is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        super(MobileNetV3, self).__init__()

        norm_type = 'batch_norm'
        act_type = 'hard_swish'

        # building first layer
        firstconv_output_channels = stage_params[0].in_channels[0]
        self.conv_first = ConvLayer(
            in_channels=3,
            out_channels=firstconv_output_channels,
            kernel_size=3,
            stride=2,
            norm_type=norm_type,
            act_type=act_type,
        )

        # building inverted residual blocks
        stages: List[nn.Module] = []

        lastconv_input_channels = stage_params[-1].out_channels[-1]
        lastconv_output_channels = 6 * lastconv_input_channels
        for stg_idx, stage_info in enumerate(stage_params):
            stage: List[nn.Module] = []

            for block in zip(stage_info.in_channels, stage_info.kernel_sizes, stage_info.expanded_channels,
                             stage_info.out_channels, stage_info.use_se, stage_info.act_type,
                             stage_info.stride):
                in_channels, kernel_size, hidden_channels, out_channels, use_se, act_type_b, stride = block
                dilation = 1 # Fix as a constant
                act_type_b = act_type_b.lower()
                stage.append(
                    InvertedResidual(in_channels=in_channels,
                                     hidden_channels=hidden_channels,
                                     out_channels=out_channels,
                                     kernel_size=kernel_size,
                                     stride=stride,
                                     norm_type=norm_type,
                                     act_type=act_type_b,
                                     use_se=use_se,
                                     dilation=dilation)
                )
            
            # add last conv
            if stg_idx == len(stage_params) - 1:
                stage.append(
                    ConvLayer(in_channels=lastconv_input_channels,
                              out_channels=lastconv_output_channels,
                              kernel_size=1,
                              norm_type=norm_type,
                              act_type=act_type,)
                )

            stage = nn.Sequential(*stage)
            stages.append(stage)
        
        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        self._feature_dim = lastconv_output_channels
        self._intermediate_features_dim = [s[-1].out_channels for s in self.stages[:-1]]
        self._intermediate_features_dim += [lastconv_output_channels]

        # init weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

        # Set batchnorm eps and momentum
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 0.001
                    m.momentum = 0.01
        self.apply(init_bn)

    def forward(self, x: Tensor):
        x = self.conv_first(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage in self.stages:
            x = stage(x)
            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x, )
        
        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def mobilenetv3(task, conf_model_backbone) -> MobileNetV3:
    return MobileNetV3(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Type

import torch
from torch.optim.lr_scheduler import _LRScheduler

from .cosine_lr import CosineAnnealingLRWithCustomWarmUp
from .cosine_warm_restart import CosineAnnealingWarmRestartsWithCustomWarmUp
from .multi_step_lr import MultiStepLR
from .poly_lr import PolynomialLRWithWarmUp
from .step_lr import StepLR

SCHEDULER_DICT: Dict[str, Type[_LRScheduler]] = {
    'cosine': CosineAnnealingWarmRestartsWithCustomWarmUp,
    'cosine_no_sgdr': CosineAnnealingLRWithCustomWarmUp,
    'multi_step': MultiStepLR,
    'poly': PolynomialLRWithWarmUp,
    'step': StepLR
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy

from ....utils import ModelOutput
from ....op.registry import ACTIVATION_REGISTRY
from ....op.custom import ConvLayer


class FC(nn.Module):
    def __init__(self, feature_dim: int, num_classes: int, params: DictConfig) -> None:
        super(FC, self).__init__()
        dropout_prob = params.dropout_prob
        num_layers = params.num_layers

        assert num_layers >= 1, "num_hidden_layers must be integer larger than 0"

        prev_size = feature_dim
        classifier = []
        for _ in range(num_layers - 1):
            classifier.append(nn.Linear(prev_size, params.intermediate_channels))
            classifier.append(ACTIVATION_REGISTRY[params.act_type]())
            prev_size = params.intermediate_channels
        classifier.append(nn.Dropout(p=dropout_prob))
        classifier.append(nn.Linear(prev_size, num_classes))
        self.classifier = nn.Sequential(*classifier)
        
    def forward(self, x: Union[Tensor, Proxy]) -> ModelOutput:
        x = self.classifier(x)
        return ModelOutput(pred=x)

class FCConv(nn.Module):
    def __init__(self, feature_dim: int, num_classes: int, params: DictConfig) -> None:
        super(FCConv, self).__init__()
        dropout_prob = params.dropout_prob
        num_layers = params.num_layers

        assert num_layers >= 1, "num_hidden_layers must be integer larger than 0"

        prev_size = feature_dim
        classifier = []
        for _ in range(num_layers - 1):
            classifier.append(ConvLayer(prev_size, params.intermediate_channels, 
                                        kernel_size=1, stride=1, padding=0, bias=False, 
                                        norm_type=params.norm_type, act_type=params.act_type))
            prev_size = params.intermediate_channels
        classifier.append(nn.Dropout(p=dropout_prob))
        classifier.append(ConvLayer(prev_size, num_classes,
                                    kernel_size=1, stride=1, padding=0, bias=True,
                                    use_act=False, use_norm=False))

        self.classifier = nn.Sequential(*classifier)

    def forward(self, x: Union[Tensor, Proxy]) -> ModelOutput:
        x = self.classifier(x.unsqueeze(-1).unsqueeze(-1))
        x = torch.flatten(x, 1)
        return ModelOutput(pred=x)

def fc(feature_dim, num_classes, conf_model_head, **kwargs) -> FC:
    return FC(feature_dim=feature_dim, num_classes=num_classes, params=conf_model_head.params)

def fc_conv(feature_dim, num_classes, conf_model_head, **kwargs) -> FCConv:
    return FCConv(feature_dim=feature_dim, num_classes=num_classes, params=conf_model_head.params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
import subprocess
from pathlib import Path
from typing import List, Literal, Optional, Union

import torch
from omegaconf import OmegaConf

from netspresso_trainer.inferencer_common import inference_common
from netspresso_trainer.utils.engine_utils import (
    LOG_LEVEL,
    get_gpus_from_parser_and_config,
    parse_args_netspresso,
    set_arguments,
    validate_inference_config,
)


def inference_with_yaml_impl(gpus: Optional[Union[List, int]], data: Union[Path, str], augmentation: Union[Path, str],
                         model: Union[Path, str], logging: Union[Path, str], environment: Union[Path, str], log_level: str = LOG_LEVEL):
    conf_environment = OmegaConf.load(environment).environment
    gpus = get_gpus_from_parser_and_config(gpus, conf_environment)
    assert isinstance(gpus, (list, int))

    gpu_ids_str = ','.join(map(str, gpus)) if isinstance(gpus, list) else str(gpus)
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids_str
    torch.cuda.empty_cache()  # Reinitialize CUDA to apply the change

    conf = set_arguments(data=data,
                         augmentation=augmentation,
                         model=model,
                         logging=logging,
                         environment=environment)
    config_summary = validate_inference_config(conf, gpus)

    try:
        if isinstance(gpus, int):
            inference_common(
                conf,
                task=config_summary.task,
                model_name=config_summary.model_name,
                logging_dir=config_summary.logging_dir,
                log_level=log_level
            )
        else:
            raise NotImplementedError
        return config_summary.logging_dir
    except Exception as e:
        raise e


def inference_cli() -> None:
    args_parsed = parse_args_netspresso(with_gpus=True, isTrain=False)

    logging_dir: Path = inference_with_yaml_impl(
        gpus=args_parsed.gpus,
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        logging=args_parsed.logging,
        environment=args_parsed.environment,
        log_level=args_parsed.log_level
    )

    return logging_dir



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from abc import ABC, abstractmethod
from typing import Dict, List, Literal, Optional

import torch
import torch.nn as nn
from omegaconf import DictConfig

from ..loggers.base import TrainingLogger
from ..utils.record import Timer
from .task_processors.base import BaseTaskProcessor


class BasePipeline(ABC):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
    ):
        super(BasePipeline, self).__init__()
        self.conf = conf
        self.task = task
        self.task_processor = task_processor
        self.model_name = model_name
        self.model = model
        self.logger = logger
        self.timer = timer

    @property
    def sample_input(self):
        return torch.randn((1, 3, self.conf.logging.model_save_options.sample_input_size[0], self.conf.logging.model_save_options.sample_input_size[1]))

    def log_results(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        samples: Optional[List] = None,
        losses : Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        data_stats: Optional[Dict] = None,
        learning_rate: Optional[float] = None,
        elapsed_time: Optional[float] = None,
    ):
        self.logger.log(
            prefix=prefix,
            epoch=epoch,
            samples=samples,
            losses=losses,
            metrics=metrics,
            data_stats=data_stats,
            learning_rate=learning_rate,
            elapsed_time=elapsed_time
        )

    @abstractmethod
    def save_summary(self):
        raise NotImplementedError



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from typing import Optional, List, Dict

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ...op.base_metaformer import Image2Sequence, MetaFormer, MetaFormerBlock, MetaFormerEncoder, MultiHeadAttention
from ...op.custom import ConvLayer
from ...op.registry import ACTIVATION_REGISTRY
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['mixtransformer']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']
TEMP_HIDDEN_SZIE_AS_CONSTANT = 256


class SegformerOverlapPatchEmbeddings(nn.Module):
    """Construct the overlapping patch embeddings."""

    def __init__(self, patch_size, stride, in_channels, hidden_size, layer_norm_eps):
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels,
            hidden_size,
            kernel_size=patch_size,
            stride=stride,
            padding=patch_size // 2,
        )

        self.flat = Image2Sequence()
        self.layer_norm = nn.LayerNorm(hidden_size, eps=layer_norm_eps)

    def forward(self, pixel_values):
        embeddings = self.proj(pixel_values)
        _, _, H_embed, W_embed = embeddings.size()
        # (batch_size, in_channels, H_embed, W_embed) -> (batch_size, in_channels, H_embed*W_embed) -> (batch_size, H_embed*W_embed, in_channels)
        embeddings = self.flat(embeddings)
        embeddings = self.layer_norm(embeddings)
        return embeddings, H_embed, W_embed


class SegformerDWConv(nn.Module):
    def __init__(self, intermediate_size=768):
        super().__init__()
        self.dwconv = nn.Conv2d(intermediate_size, intermediate_size, 3, 1, 1, bias=True, groups=intermediate_size)
        self.flat = Image2Sequence()

    def forward(self, hidden_states, height, width):
        B, N, C = hidden_states.size()  # N: height*width
        hidden_states = hidden_states.transpose(1, 2).view(B, C, height, width)
        hidden_states = self.dwconv(hidden_states)
        hidden_states = self.flat(hidden_states)

        return hidden_states


class SegformerMixFFN(nn.Module):
    def __init__(self, in_features, intermediate_size, hidden_dropout_prob, hidden_activation_type):
        super().__init__()
        self.dense1 = nn.Linear(in_features, intermediate_size)
        self.dwconv = SegformerDWConv(intermediate_size)
        self.intermediate_act_fn = ACTIVATION_REGISTRY[hidden_activation_type]()
        self.dense2 = nn.Linear(intermediate_size, in_features)

        self.dropout = nn.Dropout(hidden_dropout_prob)

    def forward(self, hidden_states, height, width):
        hidden_states = self.dense1(hidden_states)
        hidden_states = self.dwconv(hidden_states, height, width)
        hidden_states = self.intermediate_act_fn(hidden_states)
        hidden_states = self.dense2(hidden_states)

        hidden_states = self.dropout(hidden_states)
        return hidden_states


class SegFormerBlock(MetaFormerBlock):
    def __init__(self, hidden_size, num_attention_heads, attention_dropout_prob, sequence_reduction_ratio,
                 intermediate_ratio, hidden_dropout_prob, hidden_activation_type,
                 layer_norm_eps=1e-5):
        super().__init__(hidden_size, layer_norm_eps)
        self.layernorm_before = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.layernorm_after = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.token_mixer = MultiHeadAttention(
            hidden_size, num_attention_heads,
            attention_dropout_prob=attention_dropout_prob,
            use_qkv_bias=True,
            sequence_reduction_ratio=sequence_reduction_ratio
        )
        intermediate_size = int(hidden_size * intermediate_ratio)
        self.channel_mlp = SegformerMixFFN(hidden_size, intermediate_size, hidden_dropout_prob, hidden_activation_type)

    def forward(self, x, height, width):
        out_token_mixer = self.layernorm_before(x)
        out_token_mixer = self.token_mixer(out_token_mixer, height=height, width=width)

        out_token_mixer = out_token_mixer + x

        out_final = self.layernorm_after(out_token_mixer)
        out_final = self.channel_mlp(out_final, height=height, width=width)

        out_final = out_final + out_token_mixer

        return out_final


class SegformerEncoder(MetaFormerEncoder):
    def __init__(self, num_blocks, hidden_size,
                 num_attention_heads, attention_dropout_prob, sr_ratio,
                 intermediate_ratio, hidden_dropout_prob, hidden_activation_type, layer_norm_eps):
        super().__init__()
        # stochastic depth decay rule
        # drop_path_decays = [x.item() for x in torch.linspace(0, self.config.drop_path_rate, sum(self.config.depths))]

        self.blocks = nn.ModuleList()
        for _ in range(num_blocks):
            self.blocks.append(
                SegFormerBlock(
                    hidden_size,
                    num_attention_heads,
                    attention_dropout_prob,
                    sr_ratio,
                    intermediate_ratio,
                    hidden_dropout_prob,
                    hidden_activation_type,
                    layer_norm_eps
                )
            )

    def forward(self, x, height, width):
        for block in self.blocks:
            x = block(x, height, width)
        return x


class MixTransformer(MetaFormer):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'MixTransformer is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        super().__init__([stage.attention_chananels for stage in stage_params])
        self.task = task
        self.use_intermediate_features = self.task in ['segmentation', 'detection']

        intermediate_ratio = params.ffn_intermediate_expansion_ratio
        hidden_activation_type = params.ffn_act_type
        hidden_dropout_prob = params.ffn_dropout_prob
        attention_dropout_prob = params.attention_dropout_prob

        layer_norm_eps = 1e-6
        in_channels = 3
        
        self.encoder_modules = nn.ModuleList()
        for blocks in stage_params:
            num_blocks = blocks.num_blocks
            sr_ratios = blocks.sequence_reduction_ratio
            hidden_sizes = blocks.attention_chananels
            embedding_patch_sizes = blocks.embedding_patch_sizes
            embedding_strides = blocks.embedding_strides
            num_attention_heads = blocks.num_attention_heads

            module = nn.ModuleDict(
                {
                    'patch_embed': SegformerOverlapPatchEmbeddings(
                        embedding_patch_sizes,
                        embedding_strides,
                        in_channels,
                        hidden_sizes,
                        layer_norm_eps,
                    ),
                    'encoder': SegformerEncoder(
                        num_blocks,
                        hidden_sizes,
                        num_attention_heads,
                        attention_dropout_prob,
                        sr_ratios,
                        intermediate_ratio,
                        hidden_dropout_prob,
                        hidden_activation_type,
                        layer_norm_eps
                    ),
                    'norm': nn.LayerNorm(hidden_sizes, eps=layer_norm_eps)
                }
            )
            self.encoder_modules.append(module)

            in_channels = hidden_sizes

    def forward(self, x):
        B = x.size(0)
        all_hidden_states = () if self.use_intermediate_features else None

        for module in self.encoder_modules:
            x, H_embed, W_embed = module['patch_embed'](x)
            x = module['encoder'](x, height=H_embed, width=W_embed)
            x = module['norm'](x)

            x = x.reshape(B, H_embed, W_embed, -1).permute(0, 3, 1, 2).contiguous()

            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x,)

        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        B, C, _, _ = x.size()
        feat = torch.mean(x.reshape(B, C, -1), dim=2)
        return BackboneOutput(last_feature=feat)


def mixtransformer(task, conf_model_backbone) -> MixTransformer:
    return MixTransformer(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the Torchvision implementation of ResNet.
https://pytorch.org/vision/stable/_modules/torchvision/models/resnet.html
"""
from typing import Dict, List, Literal, Optional, Type, Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import BasicBlock, Bottleneck, ConvLayer, Pool2d
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['resnet']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']

BLOCK_FROM_LITERAL: Dict[str, Type[nn.Module]] = {
    'basicblock': BasicBlock,
    'bottleneck': Bottleneck,
}


class ResNet(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'ResNet is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST
        self.return_stage_idx = params.return_stage_idx if params.return_stage_idx else [len(stage_params) - 1]
        self.split_stem_conv = params.split_stem_conv

        super(ResNet, self).__init__()

        block: Literal['basicblock', 'bottleneck'] = params.block_type
        norm_layer: Optional[str] = params.norm_type

        # Fix as constant
        zero_init_residual: bool = False
        groups: int = 1
        width_per_group: int = 64
        expansion: Optional[int] = None

        block = BLOCK_FROM_LITERAL[block.lower()]

        if norm_layer is None:
            norm_layer = 'batch_norm'
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        self.groups = groups
        self.base_width = width_per_group

        if expansion is None:
            expansion = block.expansion

        if self.split_stem_conv:
            intermediate_stem_inplanes = self.inplanes // 2
            self.conv1 = nn.Sequential(
                ConvLayer(in_channels=3, out_channels=intermediate_stem_inplanes,
                                kernel_size=3, stride=2, padding=1,
                                bias=False, norm_type='batch_norm', act_type='relu'),
                ConvLayer(in_channels=intermediate_stem_inplanes, out_channels=intermediate_stem_inplanes,
                                kernel_size=3, stride=1, padding=1,
                                bias=False, norm_type='batch_norm', act_type='relu'),
                ConvLayer(in_channels=intermediate_stem_inplanes, out_channels=self.inplanes,
                                kernel_size=3, stride=1, padding=1,
                                bias=False, norm_type='batch_norm', act_type='relu'),
            )   
        else:
            self.conv1 = ConvLayer(in_channels=3, out_channels=self.inplanes,
                                kernel_size=7, stride=2, padding=3,
                                bias=False, norm_type='batch_norm', act_type='relu')
        self.maxpool = Pool2d(method='max', kernel_size=3, stride=2, padding=1)

        stages: List[nn.Module] = []

        first_stage = stage_params[0]
        layer = self._make_layer(block, first_stage.channels, first_stage.num_blocks, 
                                 expansion=expansion, downsample_flag=params.first_stage_shortcut_conv)
        stages.append(layer)
        for stage in stage_params[1:]:
            layer = self._make_layer(block, stage.channels, stage.num_blocks, stride=2,
                                     dilate=stage.replace_stride_with_dilation,
                                     expansion=expansion, downsample_pooling=stage.replace_stride_with_pooling)
            stages.append(layer)

        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        hidden_sizes = [stage.channels * expansion for stage in stage_params]
        self._feature_dim = hidden_sizes[-1]
        self._intermediate_features_dim = [hidden_sizes[i] for i in self.return_stage_idx]

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: Type[Union[BasicBlock, Bottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False, expansion: Optional[int] = None,
                    downsample_flag: bool = False, downsample_pooling: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if expansion is None:
            expansion = block.expansion
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * expansion or downsample_flag:
            if downsample_pooling:
                downsample = nn.Sequential(
                    Pool2d(method='avg', kernel_size=stride, stride=stride, ceil_mode=True, count_include_pad=False),
                    ConvLayer(
                        in_channels=self.inplanes, out_channels=planes * expansion,
                        kernel_size=1, stride=1, bias=False,
                        norm_type=norm_layer, use_act=False
                    )
                )
            else:
                downsample = ConvLayer(
                    in_channels=self.inplanes, out_channels=planes * expansion,
                    kernel_size=1, stride=stride, bias=False,
                    norm_type=norm_layer, use_act=False
                )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def forward(self, x: Tensor):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.maxpool(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage_idx, stage in enumerate(self.stages):
            x = stage(x)
            if self.use_intermediate_features and stage_idx in self.return_stage_idx:
                all_hidden_states = all_hidden_states + (x,)

        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        # x = self.fc(x)

        return BackboneOutput(last_feature=x)

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def resnet(task, conf_model_backbone) -> ResNet:
    """
        ResNet model from "Deep Residual Learning for Image Recognition" https://arxiv.org/pdf/1512.03385.pdf.
    """
    return ResNet(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------
from typing import Dict, Optional, List, Union

from omegaconf import DictConfig
import torch
from torch import Tensor
import torch.nn as nn
from torch.fx.proxy import Proxy

from ...op.custom import ConvLayer, ELAN, AConv, ADown, SPPELAN
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ["gelan"]
SUPPORTING_TASK = ['classification', 'detection']


class GELAN(nn.Module):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f"RepNCSP-ELAN is not supported on {self.task} task for now."
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST
        super().__init__()

        # Parameters
        stem_out_channels = params.stem_out_channels
        stem_kernel_size = params.stem_kernel_size
        stem_stride = params.stem_stride
        self.return_stage_idx = params.return_stage_idx if params.return_stage_idx else [len(stage_params) - 1]
        act_type = params.act_type.lower()
        stages = []

        self.conv_stem = ConvLayer(3, stem_out_channels, kernel_size=stem_kernel_size, stride=stem_stride,
                                   act_type=act_type)
        stages = []
        prev_channels = stem_out_channels

        for stage_param in stage_params:
            stage = []
            for idx, block_info in enumerate(stage_param):
                block_type = block_info[0].lower()
                in_channels = prev_channels
                out_channels = block_info[1]
                if block_type == "conv":
                    kernel_size = block_info[2]
                    stride = block_info[3]
                    block = ConvLayer(in_channels=in_channels, out_channels=out_channels,
                                           kernel_size=kernel_size, stride=stride, act_type=act_type)
                elif block_type == "elan":
                    part_channels = block_info[2]
                    use_identity = block_info[3]
                    block = ELAN(in_channels=in_channels, out_channels=out_channels, part_channels=part_channels,
                                 use_identity=use_identity, act_type=act_type, layer_type="basic")
                elif block_type == "sppelan":
                    hidden_channels = block_info[2]
                    block = SPPELAN(in_channels=in_channels, out_channels=out_channels, hidden_channels=hidden_channels,
                                    act_type=act_type)
                elif block_type == "repncspelan":
                    part_channels = block_info[2]
                    use_identity = block_info[3]
                    depth = block_info[4]
                    block = ELAN(in_channels=in_channels, out_channels=out_channels, part_channels=part_channels,
                                 use_identity=use_identity, act_type=act_type, layer_type="repncsp", n=depth)
                elif block_type == "aconv":
                    block = AConv(in_channels=in_channels, out_channels=out_channels, act_type=act_type)
                elif block_type == "adown":
                    block = ADown(in_channels=in_channels, out_channels=out_channels, act_type=act_type)
                else:
                    raise ValueError(f'Unknown block type: {block_type}')
                prev_channels = out_channels
                stage.append(block)
                assert len(stage) == idx + 1
            
            stages.append(stage)
        
        stages = [nn.Sequential(*stage) for stage in stages]
        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        stage_out_channels = [stage_param[-1][1] for stage_param in stage_params]
        self._intermediate_features_dim = [stage_out_channels[i] for i in self.return_stage_idx]
        self._feature_dim = stage_out_channels[-1]
        # Initialize
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)

    
    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        x = self.conv_stem(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage_idx, stage in enumerate(self.stages):
            x = stage(x)
            if self.use_intermediate_features and stage_idx in self.return_stage_idx:
                all_hidden_states = all_hidden_states + (x, )

        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)
        
        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)
    
    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def gelan(task, conf_model_backbone) -> GELAN:
    return GELAN(task, conf_model_backbone.params, conf_model_backbone.stage_params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List, Tuple

import cv2
import numpy as np
import torch

from ..dataloaders import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD


def _voc_color_map(N=256, normalized=False):
    def bitget(byteval, idx):
        return ((byteval & (1 << idx)) != 0)

    dtype = 'float32' if normalized else 'uint8'
    cmap = np.zeros((N, 3), dtype=dtype)
    for i in range(N):
        r = g = b = 0
        c = i
        for j in range(8):
            r = r | (bitget(c, 0) << 7 - j)
            g = g | (bitget(c, 1) << 7 - j)
            b = b | (bitget(c, 2) << 7 - j)
            c = c >> 3

        cmap[i] = np.array([r, g, b])

    cmap = cmap / 255 if normalized else cmap
    return cmap


class ClassificationVisualizer:
    def __init__(self, class_map, pallete=None) -> None:
        self.n = len(class_map)
        self.class_map = class_map

    def __call__(self, results: List[Tuple[np.ndarray, np.ndarray]], images=None):
        return_images = []
        for image, label, conf_score in zip(images, results["label"], results["conf_score"]):
            image = image.copy()
            class_name = self.class_map[label[0]] # Class is determined with top1 score
            conf_score = conf_score[0]
            prediction = f"{str(class_name)} {round(float(conf_score), 2)}"
            x1, y1 = 0, 0
            text_size, _ = cv2.getTextSize(prediction, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
            text_w, text_h = text_size
            image = cv2.rectangle(image, (x1, y1), (x1+text_w, y1+text_h+5), color=(0, 0, 255), thickness=-1)
            image = cv2.putText(image, prediction, (x1, y1+text_h), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1)

            return_images.append(image[np.newaxis, ...])
        return_images = np.concatenate(return_images, axis=0)
        return return_images


class DetectionVisualizer:
    def __init__(self, class_map, pallete=None):
        n = len(class_map)
        if pallete is None:
            self.cmap = _voc_color_map(n)
        else:
            self.cmap = np.array(pallete[:n], dtype=np.uint8)
        self.class_map = class_map


    def _convert(self, gray_image):
        assert len(gray_image.shape) == 2
        size = gray_image.shape
        color_image = np.zeros((3, size[0], size[1]), dtype=np.uint8)

        for label in range(0, len(self.cmap)):
            mask = (label == gray_image)
            color_image[0][mask] = self.cmap[label][0]
            color_image[1][mask] = self.cmap[label][1]
            color_image[2][mask] = self.cmap[label][2]

        # handle void
        mask = (gray_image == 255)
        color_image[0][mask] = color_image[1][mask] = color_image[2][mask] = 255

        return color_image

    def __call__(self, results: List[Tuple[np.ndarray, np.ndarray]], images=None):

        return_images = []
        for image, result in zip(images, results):
            image = image.copy()
            bbox_list, class_index_list = result
            for bbox_label, class_label in zip(bbox_list, class_index_list):
                class_name = self.class_map[class_label]

                # unnormalize depending on the visualizing image size
                x1 = int(bbox_label[0])
                y1 = int(bbox_label[1])
                x2 = int(bbox_label[2])
                y2 = int(bbox_label[3])
                conf_score = "" if len(bbox_label) <= 4 else " " + str(round(bbox_label[4], 2))
                color = self.cmap[class_label].tolist()

                image = cv2.rectangle(image, (x1, y1), (x2, y2), color=color, thickness=2)
                text_size, _ = cv2.getTextSize(f"{class_name}{conf_score}", cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
                text_w, text_h = text_size
                image = cv2.rectangle(image, (x1, y1-5-text_h), (x1+text_w, y1), color=color, thickness=-1)
                image = cv2.putText(image, f"{class_name}{conf_score}", (x1, y1-5), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1)

            return_images.append(image[np.newaxis, ...])
        return_images = np.concatenate(return_images, axis=0)
        return return_images


class SegmentationVisualizer:
    def __init__(self, class_map, pallete=None):
        n = len(class_map)
        if pallete is None:
            self.cmap = _voc_color_map(n)
        else:
            self.cmap = np.array(pallete[:n], dtype=np.uint8)
        self.cmap = torch.from_numpy(self.cmap[:n])

    def _convert(self, gray_image):
        assert len(gray_image.shape) == 2
        size = gray_image.shape
        color_image = np.zeros((3, size[0], size[1]), dtype=np.uint8)

        for label in range(0, len(self.cmap)):
            mask = (label == gray_image)
            color_image[0][mask] = self.cmap[label][0]
            color_image[1][mask] = self.cmap[label][1]
            color_image[2][mask] = self.cmap[label][2]

        # handle void
        mask = (gray_image == 255)
        color_image[0][mask] = color_image[1][mask] = color_image[2][mask] = 255

        return color_image

    def __call__(self, results, images=None):
        if len(results.shape) == 3:
            result_images = []
            for _real_gray_image in results:
                result_images.append(self._convert(_real_gray_image)[np.newaxis, ...])

            return np.concatenate(result_images, axis=0)
        elif len(results.shape) == 2:
            return self._convert(results)
        else:
            raise IndexError(f"gray_image.shape should be either 2 or 3, but {results.shape} were indexed.")


class PoseEstimationVisualizer:
    def __init__(self, class_map, pallete=None):
        len(class_map)

    def __call__(self, results, images=None):
        return_images = []
        for image, result in zip(images, results):
            image = image.copy()
            for keypoint in result:
                x = round(keypoint[0])
                y = round(keypoint[1])
                image = cv2.line(image, (x, y), (x, y), color=(0, 0, 255), thickness=5)

            return_images.append(image[np.newaxis, ...])
        return_images = np.concatenate(return_images, axis=0)
        return return_images


def _as_image_array(img: np.ndarray):
    min_, max_ = np.amin(img), np.amax(img)
    is_int_array = img.dtype in [np.uint8, np.uint16, np.int8, np.int16, np.int32, np.int64]
    try_uint8 = (min_ >= 0 and max_ <= 255)

    if is_int_array and try_uint8:
        img = img.astype(np.uint8)
    else:
        if min_ >= 0 and max_ <= 1:
            img = (img * 255.0).astype(np.uint8)
        elif min_ >= -0.5 and max_ <= 0.5:
            img = ((img + 0.5) * 255.0).astype(np.uint8)
        elif min_ >= -1 and max_ <= 1:
            img = ((img + 1) / 2.0 * 255.0).astype(np.uint8)
        else:
            # denormalize with mean and std
            img = np.clip(img * (np.array(IMAGENET_DEFAULT_STD, dtype=np.float32) * 255.0) + np.array(IMAGENET_DEFAULT_MEAN, dtype=np.float32) * 255.0, 0, 255).astype(np.uint8)

    if img.shape[-1] != 1 and img.shape[-1] != 3:
        img = np.expand_dims(np.concatenate([img[..., i] for i in range(img.shape[-1])], axis=0), -1)
    img = np.clip(img, a_min=0, a_max=255)
    return img


def magic_image_handler(img):
    if img.ndim == 3:
        img = img.transpose((1, 2, 0)) if img.shape[0] == 3 else img  # H x W x C
        return _as_image_array(img)
    elif img.ndim == 2:
        img = np.repeat(img[..., np.newaxis], 3, axis=2)
        return _as_image_array(img)
    elif img.ndim == 4:
        img_new = []
        for _img in img:
            _img = _img.transpose((1, 2, 0)) if _img.shape[0] == 3 else _img  # H x W x C
            img_new.append(_as_image_array(_img))
        return np.array(img_new)
    else:
        raise ValueError(f'img ndim is {img.ndim}, should be either 2, 3, or 4')
